# -*- coding: utf-8 -*-
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Resolution abstract implementation.
Overloaded methods: get_state, get_value, get_limits.
Implemented methods: _set_value, distance_to_resolution, resolution_to_distance.
Emitted signals: valueChanged.
Hardware object used: energy and detecor.
The detector object can be defined in the configuration file. If not, the
one set from the beamline configuration is used.
"""

import abc
import logging
from math import (
    asin,
    atan,
    sin,
    tan,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__copyright__ = """ Copyright © 2010-2022 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractResolution(AbstractMotor):
    """Abstract Resolution class"""

    __metaclass__ = abc.ABCMeta
    unit = "Å"

    def __init__(self, name):
        super().__init__(name)
        self._hwr_detector = None

    def init(self):
        """Initialisation"""
        super().init()

        self._hwr_detector = (
            self.get_object_by_role("detector") or HWR.beamline.detector
        )

        self.connect(self._hwr_detector.distance, "stateChanged", self.update_state)
        self.connect(self._hwr_detector.distance, "valueChanged", self.update_distance)
        self.connect(HWR.beamline.energy, "valueChanged", self.update_energy)
        self.connect(HWR.beamline.energy, "stateChanged", self.update_state)

        self.update_state(self.get_state())

        # initialise the limits
        self._nominal_limits = self.get_limits()

    def get_state(self):
        """Get the state of the distance motor.
        Returns:
            (enum 'HardwareRepositoryStates'): The state.
        """
        return self._hwr_detector.distance.get_state()

    def get_value(self):
        """Read the value.
        Returns:
            (float): value.
        """
        _distance = self._hwr_detector.distance.get_value()
        self._nominal_value = self.distance_to_resolution(_distance)

        return self._nominal_value

    def get_limits(self):
        """Return resolution low and high limits.

        Args:
            wavelength: Returns the limits for given wavelength if
                        passed, current wavelength is otherwised used

        Returns:
            (tuple): two floats tuple (low limit, high limit).
        """
        _low, _high = self._hwr_detector.distance.get_limits()

        return (
            self.distance_to_resolution(_low),
            self.distance_to_resolution(_high),
        )

    def get_limits_for_wavelength(self, wavelength: float):
        """Return resolution low and high limits.

        Args:
            wavelength: Returns the limits for given wavelength if
                        passed, current wavelength is otherwised used

        Returns:
            (tuple): two floats tuple (low limit, high limit).
        """
        _low, _high = self._hwr_detector.distance.get_limits()

        return (
            self.distance_to_resolution(_low, wavelength=wavelength),
            self.distance_to_resolution(_high, wavelength=wavelength),
        )

    def set_limits(self, limits):
        """Resolution limits are not settable.
           Set the detector distance limits instead
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def _set_value(self, value):
        """Move resolution to value.
        Args:
            value (float): target value [Å]
        """
        distance = self.resolution_to_distance(value)
        msg = f"Move resolution to {value} ({distance} mm)"
        logging.getLogger().info(msg)
        self._hwr_detector.distance.set_value(distance)

    def _calculate_resolution(self, radius, distance, wavelength=None):
        """Calculate the resolution as function of the detector radius and
        the distance.
        Args:
            radius (float): Detector radius [mm]
            distance (float): Distance from the sample to the detector [mm]
            wavelength (float): Wavelength [Å] (defaults to current wavelength)
        Returns:
            (float): Resolution [Å]
        """
        wavelength = wavelength or HWR.beamline.energy.get_wavelength()
        try:
            ttheta = atan(radius / distance)
            if ttheta:
                return wavelength / (2 * sin(ttheta / 2))
        except (TypeError, ZeroDivisionError):
            logging.getLogger().exception("Error while calculating resolution")
        return None

    def distance_to_resolution(self, distance=None, wavelength=None):
        """Convert distance to resolution.
        Args:
            distance (float): Distance [mm]. Defaults to current distance
            wavelength (float): Wavelength [Å] (defaults to current wavelength)
        Returns:
            (float): Resolution [Å].
        """
        distance = distance or self._hwr_detector.distance.get_value()

        return self._calculate_resolution(
            self._hwr_detector.get_radius(distance), distance, wavelength
        )

    def resolution_to_distance(self, resolution=None, wavelength=None):
        """Convert resolution to distance.
        Args:
            resolution(float): Resolution [Å]. Defaults to nominal value
            wavelength (float): Wavelength [Å] (defaults to current wavelength)
        Returns:
            (float): distance [mm].
        """
        resolution = resolution or self._nominal_value
        wavelength = wavelength or HWR.beamline.energy.get_wavelength()

        try:
            distance = self._hwr_detector.get_radius() / (
                tan(2 * asin(wavelength / (2 * resolution)))
            )
            return round(
                self._hwr_detector.get_radius(distance)
                / (tan(2 * asin(wavelength / (2 * resolution)))),
                2,
            )
        except (KeyError, ZeroDivisionError):
            return None

    def get_value_at_corner(self):
        """Get the resolution at the corners of the detector.
        Returns:
            (float): Resolution [Å]
        """
        _distance = self._hwr_detector.distance.get_value()
        corner_distance = self._hwr_detector.get_outer_radius()
        return self._calculate_resolution(corner_distance, _distance)

    def update_distance(self, value=None):
        """Update the resolution when distance changed.
        Args:
            value (float): Detector distance [mm].
        """
        value = value or self._hwr_detector.distance.get_value()
        self._nominal_value = self.distance_to_resolution(value)
        self.emit("valueChanged", (self._nominal_value,))

    def update_energy(self, value):
        """Update the resolution when energy changed.
        Args:
            value(float): Energy [keV]
        """
        value = value or HWR.beamline.energy.get_value()
        _wavelength = HWR.beamline.energy.calculate_wavelength(value)
        _distance = self._hwr_detector.distance.get_value()
        _radius = self._hwr_detector.get_radius(_distance)
        try:
            ttheta = atan(_radius / _distance)
            if ttheta:
                self._nominal_value = _wavelength / (2 * sin(ttheta / 2))
                self.emit("valueChanged", (self._nominal_value,))
            # finally update the limits as wavelength changed
            self.update_limits()
        except (TypeError, ZeroDivisionError):
            logging.getLogger().exception("Error while calculating resolution")

    def abort(self):
        """Abort the distance motor movement"""
        self._hwr_detector.distance.abort()

    def stop(self):
        """Stop the distance motor movement"""
        self._hwr_detector.distance.stop()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"
__credits__ = ["DESY P11"]
__category__ = "General"

import numpy as np

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractBeam import (
    AbstractBeam,
    BeamShape,
)


class P11Beam(AbstractBeam):
    def init(self):
        self.pinhole_hwobj = self.get_object_by_role("pinhole")
        self._beam_position_on_screen = [340, 256]

        self.focus_sizes = {
            -1: {"label": "unknown", "size": [0.2, 0.2]},
            0: {"label": "flat", "size": [0.2, 0.2]},
            1: {"label": "200x200", "size": [0.2, 0.2]},
            2: {"label": "100x100", "size": [0.1, 0.1]},
            3: {"label": "50x50", "size": [0.05, 0.05]},
            4: {"label": "20x20", "size": [0.02, 0.02]},
            5: {"label": "4x9", "size": [0.009, 0.004]},
        }

        self.mirror_idx_ch = self.get_channel_object("beamsize")
        self.mirror_state_ch = self.get_channel_object("state")

        if self.mirror_idx_ch is not None:
            self.mirror_idx_ch.connect_signal("update", self.mirror_idx_changed)
        if self.mirror_state_ch is not None:
            self.mirror_state_ch.connect_signal("update", self.mirror_state_changed)

        self.mirror_idx_changed()
        self.mirror_state_changed()

    def get_available_size(self):
        """Returns available beam sizes based on the current configuration."""
        return {"type": ["focus"], "values": [self.focus_sizes]}

    def get_defined_beam_size(self):
        """Implements the abstract method to return defined beam sizes."""
        return {
            "label": [item["label"] for item in self.focus_sizes.values()],
            "size": [item["size"] for item in self.focus_sizes.values()],
        }

    def set_value(self, size=None):
        """Implements the abstract method to set the beam size."""
        if isinstance(size, list):
            self._beam_width, self._beam_height = size
        elif isinstance(size, str):
            matching_size = next(
                (v for k, v in self.focus_sizes.items() if v["label"] == size), None
            )
            if matching_size:
                self._beam_width, self._beam_height = matching_size["size"]
        self.evaluate_beam_info()

    def set_beam_position_on_screen(self, beam_x_y):
        """Sets the beam position on the screen."""
        self._beam_position_on_screen = beam_x_y
        self.re_emit_values()

    def get_beam_info_state(self):
        if self.mirror_state_ch is not None:
            tango_state = self.mirror_state_ch.get_value()
            return self._convert_tango_state(tango_state)

        return self.STATES.READY

    def get_slits_gap(self):
        return None, None

    def mirror_state_changed(self, state=None):
        if state is None:
            state = self.get_beam_info_state()

        self.update_state(self._convert_tango_state(state))

    def _convert_tango_state(self, state):
        str_state = str(state)

        if str_state == "ON":
            _state = self.STATES.READY
        elif str_state == "MOVING":
            _state = self.STATES.BUSY
        else:
            _state = self.STATES.FAULT
        return _state

    def mirror_idx_changed(self, value=None):
        if value is None:
            value = self.mirror_idx_ch.get_value()

        self.log.debug(f"P11Beam - mirror idx changed. now is {value}")

        if value not in self.focus_sizes:
            value = -1
            self.log.debug(f"    - UNKNOWN mirror index")

        curr_size_item = self.focus_sizes[value]
        self.log.debug(
            f"    current mirror focus is {curr_size_item['label']}: {curr_size_item['size']}"
        )

        self._beam_size_dict["aperture"] = curr_size_item["size"]

        self.evaluate_beam_info()
        self.re_emit_values()

    def get_pinhole_size(self):
        # Keep it default as the pinhole and beamsize interaction is locked for now
        return 200

    def get_beam_focus_label(self):
        value = self.mirror_idx_ch.get_value()

        if value not in self.focus_sizes:
            value = -1
            return "UNKNOWN mirror index"
        else:
            curr_size_item = self.focus_sizes[value]
            self.log.debug(
                f"    current mirror focus is {curr_size_item['label']}: {curr_size_item['size']}"
            )

            return curr_size_item["label"]



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
ALBACollect
"""

import logging
import os
import sys
import time

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractCollect import AbstractCollect
from mxcubecore.TaskUtils import task

__author__ = "Vicente Rey Bakaikoa"
__credits__ = ["MXCuBE collaboration"]
__version__ = "2.2."


class ALBACollect(AbstractCollect):
    """Main data collection class. Inherited from AbstractMulticollect
    Collection is done by setting collection parameters and
    executing collect command
    """

    def __init__(self, name):
        """

        :param name: name of the object
        :type name: string
        """

        AbstractCollect.__init__(self, name)
        #        HardwareObject.__init__(self, name)

        self._error_msg = ""
        self.owner = None
        self.osc_id = None
        self._collecting = None

        self.helical_positions = None
        self.saved_omega_velocity = None

    def init(self):
        """
        Init method
        """

        self.ready_event = gevent.event.Event()

        self.supervisor_hwobj = self.get_object_by_role("supervisor")

        self.slowshut_hwobj = self.get_object_by_role("slow_shutter")
        self.photonshut_hwobj = self.get_object_by_role("photon_shutter")
        self.frontend_hwobj = self.get_object_by_role("frontend")

        self.ni_conf_cmd = self.get_command_object("ni_configure")
        self.ni_unconf_cmd = self.get_command_object("ni_unconfigure")

        # some extra reading channels to be saved in image header
        self.kappapos_chan = self.get_channel_object("kappapos")
        self.phipos_chan = self.get_channel_object("phipos")

        undulators = []
        try:
            for undulator in self["undulators"]:
                undulators.append(undulator)
        except Exception:
            pass

        self.exp_type_dict = {"Mesh": "raster", "Helical": "Helical"}

        det_px, det_py = HWR.beamline.detector.get_pixel_size()
        beam_div_hor, beam_div_ver = HWR.beamline.beam.get_beam_divergence()

        self.set_beamline_configuration(
            synchrotron_name="ALBA",
            directory_prefix=self.get_property("directory_prefix"),
            default_exposure_time=HWR.beamline.detector.get_default_exposure_time(),
            minimum_exposure_time=HWR.beamline.detector.get_minimum_exposure_time(),
            detector_fileext=HWR.beamline.detector.get_file_suffix(),
            detector_type=HWR.beamline.detector.get_detector_type(),
            detector_manufacturer=HWR.beamline.detector.get_manufacturer(),
            detector_model=HWR.beamline.detector.get_model(),
            detector_px=det_px,
            detector_py=det_py,
            undulators=undulators,
            focusing_optic=self.get_property("focusing_optic"),
            monochromator_type=self.get_property("monochromator"),
            beam_divergence_vertical=beam_div_ver,
            beam_divergence_horizontal=beam_div_hor,
            polarisation=self.get_property("polarisation"),
            input_files_server=self.get_property("input_files_server"),
        )

        self.emit("collectConnected", (True,))
        self.emit("collectReady", (True,))

    def data_collection_hook(self):
        """Main collection hook"""

        logging.getLogger("HWR").info("Running ALBA data collection hook")

        logging.getLogger("HWR").info("  -- wait for devices to finish moving --")
        logging.getLogger("HWR").info("       + wait for resolution...")
        HWR.beamline.resolution.wait_end_of_move()
        logging.getLogger("HWR").info("       + wait for detector distance...")
        HWR.beamline.detector.wait_move_distance_done()
        logging.getLogger("HWR").info("       + wait for energy...")
        HWR.beamline.energy.wait_move_energy_done()

        if self.aborted_by_user:
            self.emit_collection_failed("Aborted by user")
            self.aborted_by_user = False
            return

        ### EDNA_REF, OSC, MESH, HELICAL

        exp_type = self.current_dc_parameters["experiment_type"]
        logging.getLogger("HWR").debug("Running a collect (exp_type=%s)" % exp_type)

        if exp_type == "Characterization":
            logging.getLogger("HWR").debug("Running a collect (CHARACTERIZATION)")
        elif exp_type == "Helical":
            logging.getLogger("HWR").debug("Running a helical collection")
            logging.getLogger("HWR").debug(
                "   helical positions are: %s" % str(self.helical_positions)
            )
            hpos = self.helical_positions
            logging.getLogger("HWR").debug(
                "               phiy from %3.4f to %3.4f" % (hpos[0], hpos[4])
            )
            logging.getLogger("HWR").debug(
                "               phiz from %3.4f to %3.4f" % (hpos[1], hpos[5])
            )
            logging.getLogger("HWR").debug(
                "              sampx from %3.4f to %3.4f" % (hpos[2], hpos[6])
            )
            logging.getLogger("HWR").debug(
                "              sampy from %3.4f to %3.4f" % (hpos[3], hpos[7])
            )
        elif exp_type == "Mesh":
            logging.getLogger("HWR").debug("Running a raster collection ()")
            logging.getLogger("HWR").debug(
                "   number of lines are: %s" % self.mesh_num_lines
            )
            logging.getLogger("HWR").debug(
                "   total nb of frames: %s" % self.mesh_total_nb_frames
            )
            logging.getLogger("HWR").debug(
                "          mesh range : %s" % self.mesh_range
            )
            logging.getLogger("HWR").debug(
                "          mesh center : %s" % self.mesh_center
            )
        else:
            logging.getLogger("HWR").debug("Running a collect (STANDARD)")

        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]

        image_range = osc_seq["range"]
        nb_images = osc_seq["number_of_images"]
        total_range = image_range * nb_images

        ready = self.prepare_acquisition()

        if not ready:
            self.collection_failed("Cannot prepare collection")
            self.stop_collect()
            return

        self._collecting = True
        # for progressBar brick
        self.emit("progressInit", "Collection", osc_seq["number_of_images"])

        omega_pos = osc_seq["start"]

        logging.getLogger("HWR").info("Starting detector")
        self.emit("collectStarted", (self.owner, 1))

        first_image_no = osc_seq["start_image_number"]

        if exp_type == "OSC" or (exp_type == "Characterization" and nb_images == 1):
            final_pos = self.prepare_collection(
                start_angle=omega_pos,
                nb_images=nb_images,
                first_image_no=first_image_no,
            )
            HWR.beamline.detector.start_collection()
            self.collect_images(final_pos, nb_images, first_image_no)
        elif exp_type == "Characterization" and nb_images > 1:  # image one by one
            for imgno in range(nb_images):
                final_pos = self.prepare_collection(
                    start_angle=omega_pos, nb_images=1, first_image_no=first_image_no
                )
                HWR.beamline.detector.start_collection()
                self.collect_images(final_pos, 1, first_image_no)
                first_image_no += 1
                omega_pos += 90

    def collect_images(self, final_pos, nb_images, first_image_no):
        #
        # Run
        #
        logging.getLogger("HWR").info(
            "collecting images, by moving omega to %s" % final_pos
        )
        HWR.beamline.diffractometer.omega.set_value(final_pos)
        self.wait_collection_done(nb_images, first_image_no)
        self.data_collection_end()
        self.collection_finished()

    def data_collection_end(self):
        HWR.beamline.fast_shutter.cmdOut()
        HWR.beamline.diffractometer.omega.set_velocity(60)
        self.unconfigure_ni()

    def data_collection_failed(self):
        logging.getLogger("HWR").info(
            "Data collection failed. recovering sequence should go here"
        )

    def prepare_acquisition(self):
        fileinfo = self.current_dc_parameters["fileinfo"]

        basedir = fileinfo["directory"]

        #  save omega velocity
        self.saved_omega_velocity = HWR.beamline.diffractometer.omega.get_velocity()

        # create directories if needed
        self.check_directory(basedir)

        # check fast shutter closed. others opened
        shutok = self.check_shutters()

        if not shutok:
            logging.getLogger("user_level_log").error(
                " Shutters are not ready. BYPASSED. Comment line in ALBACollect.py"
            )
        else:
            logging.getLogger("user_level_log").error(
                " Shutters ready but code is BYPASSED. Comment line in ALBACollect.py"
            )

        shutok = True  # DELETE THIS AFTER TESTS

        if not shutok:
            logging.getLogger("user_level_log").error(" Shutters not ready")
            return False

        gevent.sleep(1)
        logging.getLogger("HWR").info(
            " Waiting for diffractometer to be ready. Now %s"
            % str(HWR.beamline.diffractometer.current_state)
        )
        HWR.beamline.diffractometer.wait_device_ready(timeout=10)
        logging.getLogger("HWR").info("             diffractometer is now ready.")

        # go to collect phase
        if not self.is_collect_phase():
            logging.getLogger("HWR").info(
                " Not in collect phase. Asking supervisor to go"
            )
            logging.getLogger("HWR").info(
                "  diffractometer is now ready. Now %s"
                % str(HWR.beamline.diffractometer.current_state)
            )
            success = self.go_to_collect()
            if not success:
                logging.getLogger("user_level_log").error(
                    "Cannot set COLLECT phase for diffractometer"
                )
                return False

        detok = HWR.beamline.detector.prepare_acquisition(self.current_dc_parameters)

        return detok

    def prepare_collection(self, start_angle, nb_images, first_image_no):
        # move omega to start angle
        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]

        # start_angle = osc_seq['start']
        # nb_images = osc_seq['number_of_images']

        img_range = osc_seq["range"]
        exp_time = osc_seq["exposure_time"]

        total_dist = nb_images * img_range
        total_time = nb_images * exp_time
        omega_speed = float(total_dist / total_time)

        logging.getLogger("HWR").info("  prepare detector  was not ok.")
        self.write_image_headers(start_angle)

        logging.getLogger("HWR").info(
            "  nb_images: %s / img_range: %s / exp_time: %s / total_distance: %s / speed: %s"
            % (nb_images, img_range, exp_time, total_dist, omega_speed)
        )
        logging.getLogger("HWR").info(
            "  setting omega velocity to 60 to go to intial position"
        )
        HWR.beamline.diffractometer.omega.set_velocity(60)

        omega_acceltime = HWR.beamline.diffractometer.omega.get_acceleration()

        safe_delta = 9.0 * omega_speed * omega_acceltime

        init_pos = start_angle - safe_delta
        final_pos = start_angle + total_dist + safe_delta

        logging.getLogger("HWR").info("Moving omega to initial position %s" % init_pos)
        HWR.beamline.diffractometer.omega.set_value(init_pos)

        HWR.beamline.detector.prepare_collection(nb_images, first_image_no)

        HWR.beamline.diffractometer.omega.wait_end_of_move(timeout=10)

        logging.getLogger("HWR").info(
            "Moving omega finished at %s"
            % HWR.beamline.diffractometer.omega.get_value()
        )

        # program omega speed depending on exposure time

        logging.getLogger("HWR").info("Setting omega velocity to %s" % omega_speed)
        HWR.beamline.diffractometer.omega.set_velocity(omega_speed)
        if omega_speed != 0:
            self.configure_ni(start_angle, total_dist)

        return final_pos

    def write_image_headers(self, start_angle):
        fileinfo = self.current_dc_parameters["fileinfo"]
        basedir = fileinfo["directory"]

        exp_type = self.current_dc_parameters["experiment_type"]
        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]

        nb_images = osc_seq["number_of_images"]
        # start_angle = osc_seq['start']

        img_range = osc_seq["range"]

        if exp_type == "Characterization":
            angle_spacing = 90
        else:
            angle_spacing = img_range

        exp_time = osc_seq["exposure_time"]

        # PROGRAM Image Headers
        # latency_time = 0.003
        latency_time = HWR.beamline.detector.get_latency_time()
        limaexpt = exp_time - latency_time

        self.image_headers = {}

        angle_info = [start_angle, img_range, angle_spacing]

        self.image_headers["nb_images"] = nb_images
        self.image_headers["Exposure_time"] = "%.4f" % limaexpt
        self.image_headers["Exposure_period"] = "%.4f" % exp_time
        self.image_headers["Start_angle"] = "%f deg." % start_angle
        self.image_headers["Angle_increment"] = "%f deg." % img_range
        self.image_headers["Wavelength"] = HWR.beamline.energy.get_wavelength()

        self.image_headers["Detector_distance"] = "%.5f m" % (
            HWR.beamline.detector.distance.get_value() / 1000.0
        )
        self.image_headers["Detector_Voffset"] = "0 m"

        # NBNB TODO check if get_beam_position() (should be pixels)
        # matches beamx, beamy (should be mm)
        beamx, beamy = HWR.beamline.detector.get_beam_position()
        self.image_headers["Beam_xy"] = "(%.2f, %.2f) pixels" % (beamx, beamy)

        self.image_headers["Filter_transmission"] = "%.4f" % (
            HWR.beamline.transmission.get_value() / 100.0
        )
        self.image_headers["Flux"] = "%.4g" % HWR.beamline.flux.get_value()
        self.image_headers["Detector_2theta"] = "0.0000"
        self.image_headers["Polarization"] = "0.99"
        self.image_headers["Alpha"] = "0 deg."

        self.image_headers["Kappa"] = "%.4f deg." % self.kappapos_chan.get_value()
        self.image_headers["Phi"] = "%.4f deg." % self.phipos_chan.get_value()

        self.image_headers["Chi"] = "0 deg."
        self.image_headers["Oscillation_axis"] = "X, CW"
        self.image_headers["N_oscillations"] = "1"
        self.image_headers["Detector_2theta"] = "0.0000 deg"

        self.image_headers["Image_path"] = ": %s" % basedir

        self.image_headers["Threshold_setting"] = (
            "%0f eV" % HWR.beamline.detector.get_threshold()
        )
        self.image_headers["Gain_setting"] = "%s" % str(
            HWR.beamline.detector.get_threshold_gain()
        )

        self.image_headers["Tau"] = "%s s" % str(199.1e-09)
        self.image_headers["Count_cutoff"] = "%s counts" % str(370913)
        self.image_headers["N_excluded_pixels"] = "= %s" % str(1178)
        self.image_headers["Excluded_pixels"] = ": %s" % str("badpix_mask.tif")
        self.image_headers["Trim_file"] = ": %s" % str(
            "p6m0108_E12661_T6330_vrf_m0p20.bin"
        )

        HWR.beamline.detector.set_image_headers(self.image_headers, angle_info)

    def wait_collection_done(self, nb_images, first_image_no):
        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]

        # first_image_no = osc_seq['start_image_number']
        # nb_images = osc_seq['number_of_images']
        last_image_no = first_image_no + nb_images - 1

        if nb_images > 1:
            self.wait_save_image(first_image_no)
        HWR.beamline.diffractometer.omega.wait_end_of_move(timeout=720)
        self.wait_save_image(last_image_no)

    def wait_save_image(self, frame_number, timeout=25):
        fileinfo = self.current_dc_parameters["fileinfo"]
        basedir = fileinfo["directory"]
        template = fileinfo["template"]

        filename = template % frame_number
        fullpath = os.path.join(basedir, filename)

        start_wait = time.time()

        logging.getLogger("HWR").debug("   waiting for image on disk: %s", fullpath)

        while not os.path.exists(fullpath):
            dirlist = os.listdir(basedir)  # forces directory flush ?
            if (time.time() - start_wait) > timeout:
                logging.getLogger("HWR").debug("   giving up waiting for image")
                return False
            time.sleep(0.2)

        self.last_saved_image = fullpath

        # generate thumbnails
        archive_dir = fileinfo["archive_directory"]
        self.check_directory(archive_dir)

        jpeg_filename = os.path.splitext(filename)[0] + ".jpeg"
        thumb_filename = os.path.splitext(filename)[0] + ".thumb.jpeg"

        thumb_fullpath = os.path.join(archive_dir, thumb_filename)
        jpeg_fullpath = os.path.join(archive_dir, jpeg_filename)

        logging.getLogger("HWR").debug(
            "   creating thumbnails for  %s in: %s and %s"
            % (fullpath, jpeg_fullpath, thumb_fullpath)
        )
        cmd = "adxv_thumb 0.4 %s %s" % (fullpath, jpeg_fullpath)
        os.system(cmd)
        cmd = "adxv_thumb 0.1 %s %s" % (fullpath, thumb_fullpath)
        os.system(cmd)

        logging.getLogger("HWR").debug("   writing thumbnails info in LIMS")
        self._store_image_in_lims(frame_number)

        return True

    def check_shutters(self):
        # Check fast shutter
        if HWR.beamline.fast_shutter.get_state() != 0:
            return False

        # Check slow shutter
        if self.slowshut_hwobj.get_state() != 1:
            return False

        # Check photon shutter
        if self.photonshut_hwobj.get_state() != 1:
            return False

        # Check front end
        if self.frontend_hwobj.get_state() != 1:
            return False

        return True

    def get_image_headers(self):
        headers = []
        return headers

    def collection_end(self):
        #
        # data collection end (or abort)
        #
        logging.getLogger("HWR").info(" finishing data collection ")
        HWR.beamline.fast_shutter.cmdOut()
        self.emit("progressStop")

    def check_directory(self, basedir):
        if not os.path.exists(basedir):
            try:
                os.makedirs(basedir)
            except OSError as e:
                import errno

                if e.errno != errno.EEXIST:
                    raise

    def collect_finished(self, green):
        logging.info("Data collection finished")

    def collect_failed(self, par):
        logging.exception("Data collection failed")
        self.current_dc_parameters["status"] = "failed"
        exc_type, exc_value, exc_tb = sys.exc_info()
        failed_msg = "Data collection failed!\n%s" % exc_value
        self.emit(
            "collectOscillationFailed",
            (
                self.owner,
                False,
                failed_msg,
                self.current_dc_parameters.get("collection_id"),
                1,
            ),
        )

        HWR.beamline.detector.stop_collection()
        HWR.beamline.diffractometer.omega.stop()
        self.data_collection_end()

    def go_to_collect(self, timeout=180):
        logging.getLogger("HWR").debug("sending supervisor to collect phase")
        self.supervisor_hwobj.go_collect()
        logging.getLogger("HWR").debug("supervisor sent to collect phase")

        gevent.sleep(0.5)

        t0 = time.time()
        while True:
            super_state = str(self.supervisor_hwobj.get_state()).upper()
            cphase = self.supervisor_hwobj.get_current_phase().upper()
            if super_state != "MOVING" and cphase == "COLLECT":
                break
            if time.time() - t0 > timeout:
                logging.getLogger("HWR").debug(
                    "timeout sending supervisor to collect phase"
                )
                break
            gevent.sleep(0.5)

        logging.getLogger("HWR").debug(
            "supervisor finished go collect phase task. phase is now: %s" % cphase
        )

        return self.is_collect_phase()

    def is_collect_phase(self):
        return self.supervisor_hwobj.get_current_phase().upper() == "COLLECT"

    def go_to_sampleview(self, timeout=180):
        logging.getLogger("HWR").debug("sending supervisor to sample view phase")
        self.supervisor_hwobj.go_sample_view()
        logging.getLogger("HWR").debug("supervisor sent to sample view phase")

        gevent.sleep(0.5)

        t0 = time.time()
        while True:
            super_state = str(self.supervisor_hwobj.get_state()).upper()
            cphase = self.supervisor_hwobj.get_current_phase().upper()
            if super_state != "MOVING" and cphase == "SAMPLE":
                break
            if time.time() - t0 > timeout:
                logging.getLogger("HWR").debug(
                    "timeout sending supervisor to sample view phase"
                )
                break
            gevent.sleep(0.5)

        logging.getLogger("HWR").debug(
            "supervisor finished go sample view phase task. phase is now: %s" % cphase
        )

        return self.is_sampleview_phase()

    def is_sampleview_phase(self):
        return self.supervisor_hwobj.get_current_phase().upper() == "SAMPLE"

    def configure_ni(self, startang, total_dist):
        logging.getLogger("HWR").debug(
            "Configuring NI660 with pars 0, %s, %s, 0, 1" % (startang, total_dist)
        )
        self.ni_conf_cmd(0.0, startang, total_dist, 0, 1)

    def unconfigure_ni(self):
        self.ni_unconf_cmd()

    def open_safety_shutter(self):
        """implements prepare_shutters in collect macro"""

        # prepare ALL shutters

        # close fast shutter
        if HWR.beamline.fast_shutter.get_state() != 0:
            HWR.beamline.fast_shutter.close()

        # open slow shutter
        if self.slowshut_hwobj.get_state() != 1:
            self.slowshut_hwobj.open()

        # open photon shutter
        if self.photonshut_hwobj.get_state() != 1:
            self.photonshut_hwobj.open()

        # open front end
        if self.frontend_hwobj.get_state() != 0:
            self.frontend_hwobj.open()

    def open_detector_cover(self):
        self.supervisor_hwobj.open_detector_cover()

    def open_fast_shutter(self):
        # HWR.beamline.fast_shutter.open()
        #   this function is empty for ALBA. we are not opening the fast shutter.
        #   on the contrary open_safety_shutter (equivalent to prepare_shutters in original
        #   collect macro will first close the fast shutter and open the other three
        pass

    def close_fast_shutter(self):
        HWR.beamline.fast_shutter.cmdOut()

    def close_safety_shutter(self):
        #  we will not close safety shutter during collections
        pass

    def close_detector_cover(self):
        #  we will not close detcover during collections
        #  self.supervisor.close_detector_cover()
        pass

    def set_helical_pos(self, arg):
        """
        Descript. : 8 floats describe
        p1AlignmY, p1AlignmZ, p1CentrX, p1CentrY
        p2AlignmY, p2AlignmZ, p2CentrX, p2CentrY
        """
        self.helical_positions = [
            arg["1"]["phiy"],
            arg["1"]["phiz"],
            arg["1"]["sampx"],
            arg["1"]["sampy"],
            arg["2"]["phiy"],
            arg["2"]["phiz"],
            arg["2"]["sampx"],
            arg["2"]["sampy"],
        ]

    def setMeshScanParameters(self, num_lines, num_images_per_line, mesh_range):
        """
        Descript. :
        """
        pass

    @task
    def _take_crystal_snapshot(self, filename):
        """
        Descript. :
        """
        if not self.is_sampleview_phase():
            self.go_to_sampleview()

        HWR.beamline.sample_view.save_snapshot(filename)
        logging.getLogger("HWR").debug(" - snapshot saved to %s" % filename)

    @task
    def move_motors(self, motor_position_dict):
        """
        Descript. :
        """
        HWR.beamline.diffractometer.move_motors(motor_position_dict)

    def create_file_directories(self):
        """
        Method create directories for raw files and processing files.
        Directorie for xds.input and auto_processing are created
        """
        self.create_directories(
            self.current_dc_parameters["fileinfo"]["directory"],
            self.current_dc_parameters["fileinfo"]["process_directory"],
        )

        """create processing directories and img links"""
        xds_directory, auto_directory, ednaproc_directory = self.prepare_input_files()

        try:
            self.create_directories(xds_directory, auto_directory, ednaproc_directory)
            os.system(
                "chmod -R 777 %s %s %s"
                % (xds_directory, auto_directory, ednaproc_directory)
            )
            """todo, create link of imgs for auto_processing
            try:
                os.symlink(files_directory, os.path.join(process_directory, "img"))
            except os.error, e:
                if e.errno != errno.EEXIST:
                    raise
            """
            # os.symlink(files_directory, os.path.join(process_directory, "img"))
        except Exception:
            logging.exception("Could not create processing file directory")
            return

        if xds_directory:
            self.current_dc_parameters["xds_dir"] = xds_directory

        if auto_directory:
            self.current_dc_parameters["auto_dir"] = auto_directory

        # pass wavelength needed in auto processing input files

        osc_pars = self.current_dc_parameters["oscillation_sequence"][0]
        osc_pars["wavelength"] = HWR.beamline.energy.get_wavelength()

        HWR.beamline.offline_processing.create_input_files(
            xds_directory, auto_directory, self.current_dc_parameters
        )

    def prepare_input_files(self):
        """
        Descript. :
        """
        i = 1
        log = logging.getLogger("user_level_log")

        while True:
            xds_input_file_dirname = "xds_%s_%s_%d" % (
                self.current_dc_parameters["fileinfo"]["prefix"],
                self.current_dc_parameters["fileinfo"]["run_number"],
                i,
            )
            xds_directory = os.path.join(
                self.current_dc_parameters["fileinfo"]["process_directory"],
                xds_input_file_dirname,
            )
            if not os.path.exists(xds_directory):
                break
            i += 1

        self.current_dc_parameters["xds_dir"] = xds_directory

        mosflm_input_file_dirname = "mosflm_%s_%s_%d" % (
            self.current_dc_parameters["fileinfo"]["prefix"],
            self.current_dc_parameters["fileinfo"]["run_number"],
            i,
        )
        mosflm_directory = os.path.join(
            self.current_dc_parameters["fileinfo"]["process_directory"],
            mosflm_input_file_dirname,
        )

        log.info("  - xds: %s / mosflm: %s" % (xds_directory, mosflm_directory))

        while True:
            ednaproc_dirname = "ednaproc_%s_%s_%d" % (
                self.current_dc_parameters["fileinfo"]["prefix"],
                self.current_dc_parameters["fileinfo"]["run_number"],
                i,
            )
            ednaproc_directory = os.path.join(
                self.current_dc_parameters["fileinfo"]["process_directory"],
                ednaproc_dirname,
            )
            if not os.path.exists(ednaproc_directory):
                break
            i += 1

        self.current_dc_parameters["ednaproc_dir"] = ednaproc_directory

        return xds_directory, mosflm_directory, ednaproc_directory

    def get_undulators_gaps(self):
        """
        Descript. : return triplet with gaps. In our case we have one gap,
                    others are 0
        """
        # TODO
        try:
            if self.chan_undulator_gap:
                und_gaps = self.chan_undulator_gap.get_value()
                if type(und_gaps) in (list, tuple):
                    return und_gaps
                else:
                    return und_gaps
        except Exception:
            pass
        return {}

    def get_slit_gaps(self):
        """
        Descript. :
        """
        if HWR.beamline.beam is not None:
            return HWR.beamline.beam.get_slits_gap()
        return None, None

    def get_beam_shape(self):
        """
        Descript. :
        """
        if HWR.beamline.beam is not None:
            return HWR.beamline.beam.get_beam_shape()

    def get_machine_current(self):
        """
        Descript. :
        """
        if HWR.beamline.machine_info:
            return HWR.beamline.machine_info.get_current()
        else:
            return 0

    def get_machine_message(self):
        """
        Descript. :
        """
        if HWR.beamline.machine_info:
            return HWR.beamline.machine_info.get_message()
        else:
            return ""

    def get_machine_fill_mode(self):
        """
        Descript. :
        """
        if HWR.beamline.machine_info:
            return "FillMode not/impl"
            # fill_mode = str(self.machine_info_hwobj.get_message())
            # return fill_mode[:20]
        else:
            return ""

    def trigger_auto_processing(self, event, frame):
        if event == "after":
            dc_pars = self.current_dc_parameters
            HWR.beamline.offline_processing.trigger_auto_processing(dc_pars)


def test_hwo(hwo):
    print("Shutters (ready for collect): ", hwo.check_shutters())
    print("Supervisor(collect phase): ", hwo.is_collect_phase())

    print("Kappa ", hwo.kappapos_chan.get_value())
    print("Phi ", hwo.phipos_chan.get_value())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Transmission control for MANACA.
Attenuator: ABS-300-12-C-DN-DN40 Flanges.
Foils: #1, 8 um Al;
       #2, 10 um Al;
       #3, 20 um Al;
       #4, 80 um Al;
       #5, 160 um Al;
       #6, 320 um Al;
       #7, 800 um Al;
       #8, 1500 um Al;
       #9, 8 um Ti;
       #10, 10 um Cu
       #11, 5 um Au;
       #12, 25 um Zr.
 Here we use the Beer-Lambert law to calculate the transmission/attenuation: I = I0 * e^-(MU * d)
 were I is the transmitted (final) beam intensity, I0 is the incident (initial) beam intensity,
 MU is the linear attenuation coefficient (cm^-1) and d is the thickness of the foil (cm).
"""

import argparse
import itertools
import math
import time

from epics import PV


def read_input():
    # Read inputs from terminal
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-e",
        "--energy",
        required=True,
        type=float,
        help="Energy of the x-ray beam [keV]",
    )
    ap.add_argument(
        "-t",
        "--transmission",
        required=True,
        type=float,
        help="Transmission to be set, [%]",
    )
    args = vars(ap.parse_args())
    return args


def get_transmission(energy, transmission):
    transmission = transmission / 100
    # Expressions (interpolation) for calculations of MU (linear attenuation coefficient [cm^-1]) for each material in function of the beam energy.
    # The data used to obtain the expressions (curve fitting) are from  "https://physics.nist.gov/PhysRefData/FFast/html/form.html"
    # This attenuation coefficients can be verified and optimised after experimental validation at the beamline.
    MU_Al = (
        (78657.01011 * math.exp(-energy / 0.65969))
        + (6406.36151 * math.exp(-energy / 1.63268))
        + (492.29999 * math.exp(-energy / 4.42554))
        + (3.2588)
    )
    MU_Ti = (
        (28062.17632 * math.exp(-energy / 1.88062))
        + (6354120 * math.exp(-energy / 0.49973))
        + (2257.91488 * math.exp(-energy / 5.28296))
        + (17.4342)
    )
    MU_Cu = (
        (1147850000 * math.exp(-energy / 0.56933))
        + (2582.7593 * math.exp(-energy / 8.40671))
        + (30628.08291 * math.exp(-energy / 2.98937))
        + (22.4187)
    )
    MU_Au = (42783.04258 * math.exp(-energy / 5.39244)) + (399.59714)
    MU_Zr = (11143.28458 * math.exp(-energy / 5.87029)) + (102.27474)
    MU_EMPTY = 0.0
    # Foils in ABS-attenuator. The dictionary gives the foil position and material in the key (e.g. 'F1_Al', position 1 occupied by an aluminium foil),
    # and the thickness [um] in the value. The list gives the positions and materials, similar to the dictionary, and it is used in the combination.
    foils_dict = {
        "F0_EMPTY": 0,
        "F1_Al": 8,
        "F2_Al": 10,
        "F3_Al": 20,
        "F4_Al": 80,
        "F5_Al": 160,
        "F6_Al": 320,
        "F7_Al": 800,
        "F8_Al": 1500,
        "F9_Ti": 8,
        "F10_Cu": 10,
        "F11_Au": 5,
        "F12_Zr": 25,
    }

    attenuator_position = [
        "F0_EMPTY",
        "F1_Al",
        "F2_Al",
        "F3_Al",
        "F4_Al",
        "F5_Al",
        "F6_Al",
        "F7_Al",
        "F8_Al",
        "F9_Ti",
        "F10_Cu",
        "F11_Au",
        "F12_Zr",
    ]
    # conditions to use only the continuous part of the absorption spectra and not the edge region. It can be optimised after experimental validation.
    if energy < 6.5:
        attenuator_position = attenuator_position[0:9]
    elif 6.5 <= energy < 9.5:
        attenuator_position = attenuator_position[0:10]
    elif 9.5 <= energy < 15.0:
        attenuator_position = attenuator_position[0:11]
    elif 15.0 <= energy < 18.5:
        attenuator_position = attenuator_position[0:12]
    elif energy >= 18.5:
        attenuator_position = attenuator_position[0:13]

    foils_comb = []
    atten_coef_sum = []

    # calculate the all possible unique combinations
    for i in range(1, len(attenuator_position) + 1):
        product = itertools.combinations(attenuator_position, i)
        for item in product:
            foils_comb.append(list(item))
    # calculate the product of MU (cm^-1) * thickness (cm) for each material (MU * d)
    # and sum the products to obtain the total (final) attenuation for each possible combination.
    # and put the results (total attenuation) in a list.
    comb_sum = []
    for item in foils_comb:
        comb_tmp = []
        for sub_item in item:
            if "_Al" in sub_item:
                atten_mu_x = float(foils_dict[sub_item] * 1.0e-4) * MU_Al
                comb_tmp.append(atten_mu_x)
            elif "_Ti" in sub_item:
                atten_mu_x = float(foils_dict[sub_item] * 1.0e-4) * MU_Ti
                comb_tmp.append(atten_mu_x)
            elif "_Cu" in sub_item:
                atten_mu_x = float(foils_dict[sub_item] * 1.0e-4) * MU_Cu
                comb_tmp.append(atten_mu_x)
            elif "_Au" in sub_item:
                atten_mu_x = float(foils_dict[sub_item] * 1.0e-4) * MU_Au
                comb_tmp.append(atten_mu_x)
            elif "_Zr" in sub_item:
                atten_mu_x = float(foils_dict[sub_item] * 1.0e-4) * MU_Zr
                comb_tmp.append(atten_mu_x)
            elif "_EMPTY" in sub_item:
                atten_mu_x = float(foils_dict[sub_item] * 1.0e-4) * MU_EMPTY
                comb_tmp.append(atten_mu_x)

        comb_sum.append(sum(comb_tmp))

    # create a dictionary to correlate the combinations of foils (value) with the total attenuation (key)
    FOILS = dict(zip(comb_sum, foils_comb))
    # calculate the transmission for each combination using the product calculated before (the keys of FOILS)
    FOILS_T = {1.0 * (math.exp(-k)): v for k, v in FOILS.items()}
    # find the closest transmission that is possible to obtain with the foils on ABS-300.
    closest_transmission = min(
        FOILS_T.keys(), key=lambda k: abs(float(k) - transmission)
    )
    # get the foil combination for the closest transmission found
    filter_combination = FOILS_T[closest_transmission]

    return transmission, closest_transmission, filter_combination


def set_foils(filter_combination):
    # declare a list with available foils and a dictionary with epics PV class (FOIL ACT, FOIL IN, FOIL OUT) for each foil
    attenuator_position = [
        "F1_Al",
        "F2_Al",
        "F3_Al",
        "F4_Al",
        "F5_Al",
        "F6_Al",
        "F7_Al",
        "F8_Al",
        "F9_Ti",
        "F10_Cu",
        "F11_Au",
        "F12_Zr",
    ]

    foils_pv = {
        "F1_Al": (
            PV("MNC:B:RIO01:9474C:bo0"),
            PV("MNC:B:RIO01:9425A:bi11"),
            PV("MNC:B:RIO01:9425A:bi12"),
        ),
        "F2_Al": (
            PV("MNC:B:RIO01:9474C:bo1"),
            PV("MNC:B:RIO01:9425A:bi10"),
            PV("MNC:B:RIO01:9425A:bi13"),
        ),
        "F3_Al": (
            PV("MNC:B:RIO01:9474C:bo2"),
            PV("MNC:B:RIO01:9425A:bi9"),
            PV("MNC:B:RIO01:9425A:bi14"),
        ),
        "F4_Al": (
            PV("MNC:B:RIO01:9474C:bo3"),
            PV("MNC:B:RIO01:9425A:bi8"),
            PV("MNC:B:RIO01:9425A:bi15"),
        ),
        "F5_Al": (
            PV("MNC:B:RIO01:9474C:bo4"),
            PV("MNC:B:RIO01:9425A:bi7"),
            PV("MNC:B:RIO01:9425A:bi16"),
        ),
        "F6_Al": (
            PV("MNC:B:RIO01:9474C:bo5"),
            PV("MNC:B:RIO01:9425A:bi6"),
            PV("MNC:B:RIO01:9425A:bi17"),
        ),
        "F7_Al": (
            PV("MNC:B:RIO01:9474C:bo6"),
            PV("MNC:B:RIO01:9425A:bi5"),
            PV("MNC:B:RIO01:9425A:bi18"),
        ),
        "F8_Al": (
            PV("MNC:B:RIO01:9474C:bo7"),
            PV("MNC:B:RIO01:9425A:bi4"),
            PV("MNC:B:RIO01:9425A:bi19"),
        ),
        "F9_Ti": (
            PV("MNC:B:RIO01:9474D:bo0"),
            PV("MNC:B:RIO01:9425A:bi3"),
            PV("MNC:B:RIO01:9425A:bi20"),
        ),
        "F10_Cu": (
            PV("MNC:B:RIO01:9474D:bo1"),
            PV("MNC:B:RIO01:9425A:bi2"),
            PV("MNC:B:RIO01:9425A:bi21"),
        ),
        "F11_Au": (
            PV("MNC:B:RIO01:9474D:bo2"),
            PV("MNC:B:RIO01:9425A:bi1"),
            PV("MNC:B:RIO01:9425A:bi22"),
        ),
        "F12_Zr": (
            PV("MNC:B:RIO01:9474D:bo3"),
            PV("MNC:B:RIO01:9425A:bi0"),
            PV("MNC:B:RIO01:9425A:bi23"),
        ),
    }

    # setup the foils: check foil status, put the foils in the 'filter_combination' and remove the others

    wt = 0.1

    # the commented lines below works for the attenuator expected logic (0 for foil out and 1 for foil in).
    # as the current logic is inverted (1 is foil out and 0 is foil in) use the uncommented lines.

    for foil in attenuator_position:
        if foil in filter_combination:
            # if foils_pv[foil][1].get() == 0 and foils_pv[foil][2].get() == 1:
            if foils_pv[foil][1].get() == 1 and foils_pv[foil][2].get() == 0:
                # foils_pv[foil][0].put(1)
                foils_pv[foil][0].put(0)
                status = 0
                time.sleep(wt)
            # elif foils_pv[foil][1].get() == 1 and foils_pv[foil][2].get() == 0:
            elif foils_pv[foil][1].get() == 0 and foils_pv[foil][2].get() == 1:
                status = 0
            elif foils_pv[foil][1].get() == foils_pv[foil][2].get():
                status = 1
                break
        else:
            # if foils_pv[foil][1].get() == 0 and foils_pv[foil][2].get() == 1:
            if foils_pv[foil][1].get() == 1 and foils_pv[foil][2].get() == 0:
                status = 0
            # elif foils_pv[foil][1].get() == 1 and foils_pv[foil][2].get() == 0:
            elif foils_pv[foil][1].get() == 0 and foils_pv[foil][2].get() == 1:
                # foils_pv[foil][0].put(0)
                foils_pv[foil][0].put(1)
                status = 0
                time.sleep(wt)
            elif foils_pv[foil][1].get() == foils_pv[foil][2].get():
                status = 1
                break

    return status


def main():
    energy = read_input()["energy"]
    transmission = read_input()["transmission"]
    transmission_setup = get_transmission(energy, transmission)
    # transmission required by the user
    user_transmission = transmission_setup[0] * 100
    # real transmission got with calculated foil combination
    actual_transmission = round(transmission_setup[1] * 100, 2)
    # calculated foil combination to get the required transmission
    filter_combination = transmission_setup[2]
    # put the foils and check if the they were correctly positioned (status:0 = ok; 1 = fail)
    foil_status = set_foils(filter_combination)


if __name__ == "__main__":
    main()



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""EMBLEnergy"""

import logging

import gevent

from mxcubecore.HardwareObjects.abstract.AbstractEnergy import AbstractEnergy

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLEnergy(AbstractEnergy):
    """
    Defines interface to the Tine energy server
    """

    def __init__(self, name):
        AbstractEnergy.__init__(self, name)

        self.ready_event = None
        self.undulator_gaps = ()
        self.ctrl_bytes = None
        self.bragg_break_status = None
        self.do_beam_alignment = False
        self.delta = 0
        self._tunable = True
        self._energy_value = None
        self._wavelength_value = None
        self._energy_limits = ()
        self._moving = None

        self.chan_energy = None
        self.chan_limit_low = None
        self.chan_limit_high = None
        self.chan_status = None
        self.chan_undulator_gaps = None
        self.chan_status_bragg_break = None
        self.cmd_set_energy = None
        self.cmd_energy_ctrl_byte = None
        self.cmd_set_break_bragg = None
        self.cmd_release_break_bragg = None
        self.cmd_reset_perp = None

    def init(self):
        self.ready_event = gevent.event.Event()

        self.cmd_set_energy = self.get_command_object("cmdSetEnergy")
        self.cmd_energy_ctrl_byte = self.get_command_object("cmdEnergyCtrlByte")
        self.cmd_set_break_bragg = self.get_command_object("cmdSetBreakBragg")
        self.cmd_release_break_bragg = self.get_command_object("cmdReleaseBreakBragg")
        self.cmd_reset_perp = self.get_command_object("cmdResetPerp")

        self.chan_energy = self.get_channel_object("chanEnergy")
        if self.chan_energy is not None:
            self.chan_energy.connect_signal("update", self.energy_position_changed)

        self.chan_limit_low = self.get_channel_object("chanLimitLow", optional=True)
        if self.chan_limit_low is not None:
            self.chan_limit_low.connect_signal("update", self.energy_limits_changed)

        self.chan_limit_high = self.get_channel_object("chanLimitHigh", optional=True)
        if self.chan_limit_high is not None:
            self.chan_limit_high.connect_signal("update", self.energy_limits_changed)

        self.chan_status = self.get_channel_object("chanStatus")
        if self.chan_status is not None:
            self.chan_status.connect_signal("update", self.energy_state_changed)

        self.chan_undulator_gaps = self.get_channel_object(
            "chanUndulatorGap", optional=True
        )
        if self.chan_undulator_gaps is not None:
            self.chan_undulator_gaps.connect_signal(
                "update", self.undulator_gaps_changed
            )

        self.chan_status_bragg_break = self.get_channel_object("chanStatusBraggBreak")
        if self.chan_status_bragg_break is not None:
            self.chan_status_bragg_break.connect_signal(
                "update", self.bragg_break_status_changed
            )

        try:
            self._default_energy = self.get_property("defaultEnergy")
        except Exception:
            logging.getLogger("HWR").warning("Energy: no default energy defined")

        try:
            self._energy_limits = eval(self.get_property("staticLimits"))
        except Exception:
            self._energy_limits = (None, None)
        self.ctrl_bytes = eval(self.get_property("ctrlBytes"))

        if not self.chan_energy:
            self.energy_position_changed(self._default_energy * 1000)

    def set_do_beam_alignment(self, state):
        """
        Enables/disable beam alignment after changing the energy
        :param state: boolean
        :return:
        """
        self.do_beam_alignment = state

    def get_value(self):
        """
        Returns current energy in keV
        :return: float
        """

        value = self._default_energy
        if self.chan_energy is not None:
            try:
                value = self.chan_energy.get_value()
                return value[0] / 1000
            except Exception:
                logging.getLogger("HWR").exception(
                    "Energy: could not read current energy"
                )
                return None
        return value

    def get_limits(self):
        """
        Returns energy limits as list of two floats
        :return: (float, float)
        """
        if self.chan_limit_low is not None and self.chan_limit_high is not None:
            try:
                self._energy_limits = (
                    self.chan_limit_low.get_value(),
                    self.chan_limit_high.get_value(),
                )
            except Exception:
                logging.getLogger("HWR").exception(
                    "Energy: could not read energy limits"
                )
        return self._energy_limits

    def get_wavelength_limits(self):
        """
        Returns wavelength limits as list of two floats
        :return: (float, float)
        """
        lims = None
        self._energy_limits = self.getEnergyLimits()
        if self._energy_limits is not None:
            lims = (12.3984 / self._energy_limits[1], 12.3984 / self.en_lims[0])
        return lims

    def move_energy_started(self):
        """
        Emits moveEnergyStarted signal
        :return:
        """
        self.emit("moveEnergyStarted", ())

    def move_energy_failed(self):
        """
        Emits moveEnergyFailedsignal
        :return:
        """
        self._moving = False
        self.emit("moveEnergyFailed", ())

    def move_energy_aborted(self):
        """
        Emits moveEnergyFailed signal
        :return:
        """
        self._moving = False
        self.emit("moveEnergyFailed", ())

    def move_energy_finished(self, result):
        """
        Emits moveEnergyFinished signal
        :param result:
        :return:
        """
        self._moving = False
        self.emit("moveEnergyFinished", ())

    def check_limits(self, value):
        """
        Checks given value if it is within limits
        """
        logging.getLogger("HWR").info("Checking the move limits")
        result = False

        if self._energy_limits[0] <= value <= self.en_lims[1]:
            logging.getLogger("HWR").info("Limits ok")
            result = True
        else:
            logging.getLogger("GUI").info("Energy: Requested value is out of limits")
        return result

    #
    # def start_move_wavelength(self, value, wait=True):
    #     """
    #     Starts wavelength change
    #     :param value: float
    #     :param wait: boolean
    #     :return:
    #     """
    #     logging.getLogger("HWR").info("Moving wavelength to (%s)" % value)
    #     return self.move_energy(12.3984 / value, wait)
    #     # return self.startMoveEnergy(value, wait)

    def cancel_move_energy(self):
        """
        Cancels energy change
        :return:
        """
        logging.getLogger("user_level_log").info("Energy: Cancel move")
        # self.moveEnergy.abort()

    def set_value(self, energy, wait=True):
        """
        Sets energy in keV
        """
        # gevent.spawn(self.move_energy_task(energy))
        self.move_energy_task(energy)

    def move_energy_task(self, energy):
        """
        Actual energy change task
        :param energy: in keV, float
        :return:
        """
        current_en = self.get_value()
        pos = abs(current_en - energy)
        self.delta = pos
        if pos < 0.001:
            self.emit("stateChanged", ("ready",))
        else:
            # if energy <= 6:
            #    self.cmd_energy_ctrl_byte(self.ctrl_bytes[0])
            # else:
            #    self.cmd_energy_ctrl_byte(self.ctrl_bytes[1])
            if self.cmd_energy_ctrl_byte is not None:
                if pos > 0.1:
                    # p13 63, p14 15
                    self.cmd_energy_ctrl_byte(self.ctrl_bytes[1])
                else:
                    self.cmd_energy_ctrl_byte(self.ctrl_bytes[0])

            self._moving = pos
            self.release_break_bragg()
            gevent.sleep(2)

            if self.cmd_set_energy:
                logging.getLogger("GUI").info("Energy: Moving to %.2f keV", energy)
                self.emit("statusInfoChanged", "Moving to %.2f keV" % energy)
                self.cmd_set_energy(energy)
            else:
                # Mockup mode
                self.energy_position_changed([energy * 1000])

    def energy_position_changed(self, pos):
        """
        Event called when energy value has been changed
        :param pos: float
        :return:
        """
        # self.moveEnergyCmdFinished(True)
        if isinstance(pos, (list, tuple)):
            pos = pos[0]
        value = pos / 1000

        if self._nominal_value is None or abs(value - self._nominal_value) > 1e-3:
            self.update_value(value)

    def energy_limits_changed(self, limits):
        """
        Updates energy limits
        :param limits: (float, float)
        :return:
        """
        self.update_limits(self.get_limits())

    def energy_state_changed(self, state):
        """
        Updates energy status
        :param state: int
        :return:
        """
        # logging.getLogger('HWR').info("Energy: State changed to %s" % str(state))
        self.energy_server_check_for_errors(state)
        state = int(state[0])
        if state == 0:
            if self._moving:
                self._moving = False
                self.set_break_bragg()
                if self.cmd_reset_perp is not None:
                    logging.getLogger("HWR").info("Energy: Perp reset sent")
                    self.cmd_reset_perp()
            self.move_energy_finished(0)
            self.update_state(self.STATES.READY)
            # self.emit("stateChanged", "ready")
            self.emit("statusInfoChanged", "")
            if self.do_beam_alignment and self.delta > 0.1:
                self.emit("beamAlignmentRequested")
            self.delta = 0

        elif state == 1:
            self.move_energy_started()
            self.update_state(self.STATES.BUSY)
            # self.emit("stateChanged", "busy")

    def wait_ready(self, timeout=20):
        """
        Waits till energy change is done
        :param timeout: sec in int
        :return:
        """
        super(EMBLEnergy, self).wait_ready(timeout=20)

    def energy_server_check_for_errors(self, state):
        """
        Displays error message if the energy change fails
        :param state: list of ints
        :return:
        """
        if state[0] == 1.0 or state[1] == 63:
            return
        elems = ["hdm1", "hdm2", "roll", "undulator", "bragg", "perp"]
        message = "Energy: Error, setting energy failed on motors: "
        bits = [int(state[1]) >> i & 1 for i in range(5, -1, -1)]
        # logging.getLogger('GUI').error("%s"%bits)
        for i in range(6):
            if bits[i] == 0:
                message = "%s %s" % (message, elems[i])
        logging.getLogger("GUI").error(message)
        self.emit("statusInfoChanged", message)

    def bragg_break_status_changed(self, status):
        """
        Updates status of bragg breaks
        :param status:
        :return:
        """
        self.bragg_break_status = status

    def re_emit_values(self):
        """
        Reemits signals
        :return:
        """
        self.emit("energyChanged", (self._energy_value, self._wavelength_value))
        self.emit("valueChanged", (self._energy_value,))

    def undulator_gaps_changed(self, value):
        """
        Updates undulator gaps
        :param value:
        :return:
        """
        if isinstance(value, (list, tuple)):
            self.undulator_gaps = value[0]
        else:
            self.undulator_gaps = value

    def get_undulator_gaps(self):
        """
        Returns undulator gaps
        :return:
        """
        if self.chan_undulator_gaps:
            self.undulator_gaps_changed(self.chan_undulator_gaps.get_value())
        return self.undulator_gaps

    def set_break_bragg(self):
        """
        Sets bragg breaks
        :return:
        """
        if self.chan_status_bragg_break.get_value() != 0:
            logging.getLogger("GUI").warning("Energy: Setting bragg brake...")
            self.emit("statusInfoChanged", "Setting Bragg break...")
            gevent.sleep(3)
            self.wait_ready()
            gevent.sleep(1)
            self.wait_ready()
            logging.getLogger("HWR").info("Energy: Set bragg break cmd send")
            self.cmd_set_break_bragg(1)
            gevent.sleep(2)
            if self.chan_status_bragg_break is not None:
                logging.getLogger("HWR").warning(
                    "Energy: Waiting for break set (first try) ..."
                )
                with gevent.Timeout(
                    20, Exception("Energy: Timeout waiting for break set")
                ):
                    while self.chan_status_bragg_break.get_value() != 0:
                        gevent.sleep(0.1)
                gevent.sleep(3)
                logging.getLogger("HWR").warning(
                    "Waiting for break set (second try) ..."
                )
                with gevent.Timeout(20, Exception("Timeout waiting for break set")):
                    while self.chan_status_bragg_break.get_value() != 0:
                        gevent.sleep(0.1)
            else:
                gevent.sleep(10)
            self.emit("statusInfoChanged", "Bragg break set")
            logging.getLogger("GUI").info("Energy: Bragg brake set")
        else:
            logging.getLogger("GUI").info("Energy: Bragg brake already set")

    def release_break_bragg(self):
        """
        Release bragg breaks
        :return:
        """
        if self.chan_status_bragg_break.get_value() != 1:
            logging.getLogger("GUI").warning("Energy: Releasing bragg brake...")
            self.emit("statusInfoChanged", "Releasing Bragg break...")
            self.cmd_release_break_bragg(1)
            gevent.sleep(2)
            if self.chan_status_bragg_break is not None:
                logging.getLogger("GUI").warning(
                    "Energy: Waiting for brake released..."
                )
                with gevent.Timeout(
                    20, Exception("Energy: Timeout waiting for break release")
                ):
                    while self.chan_status_bragg_break.get_value() != 1:
                        gevent.sleep(0.1)

            else:
                logging.getLogger("GUI").info(
                    "Energy: Sleep 10 sec before brake released..."
                )
                gevent.sleep(10)
            logging.getLogger("GUI").info("Energy: Bragg brake released")
            self.emit("statusInfoChanged", "Bragg break released")
        else:
            logging.getLogger("GUI").info("Energy: Bragg brake already released")
            self.emit("statusInfoChanged", "Bragg break released")



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractTransmission import (
    AbstractTransmission,
)
from mxcubecore.HardwareObjects.LNLS.EPICSActuator import EPICSActuator
from mxcubecore.HardwareObjects.LNLS.read_transmission_mnc import read_transmission
from mxcubecore.HardwareObjects.LNLS.set_transmission_mnc import (
    get_transmission,
    set_foils,
)

__copyright__ = """ Copyright © 2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class LNLSTransmission(EPICSActuator, AbstractTransmission):
    """Transmission value as a percentage"""

    def init(self):
        """Override method."""
        AbstractTransmission.init(self)
        self.energy = HWR.beamline.energy

    def get_value(self):
        """Override method."""
        try:
            energy_val = float(self.energy.get_value())  # Check if valid energy
            value, status = read_transmission(energy_val)
        except Exception as e:
            return "--"

        if status == 1:
            return "--"  # Invalid transmission read

        percentage = round(value * 100, 2)
        return percentage

    def _set_value(self, value):
        """Override method."""
        try:
            energy_val = float(self.energy.get_value())  # Check if valid energy
            _, actual_transmission, filter_setup = get_transmission(energy_val, value)
            actual_transmission = round(actual_transmission * 100, 2)

            logging.getLogger("HWR").info(
                "Requested transmission: %s. Closest possible value: %s"
                % (str(value), str(actual_transmission))
            )

            # status 0 is ok, status 1 is failure
            foil_status = set_foils(filter_setup)
        except Exception as e:
            logging.getLogger("HWR").error(
                "Error while setting transmission: %s" % str(e)
            )
        else:
            if foil_status == 0:
                logging.getLogger("HWR").info("Transmission is successfully set!")
                return
            logging.getLogger("HWR").error(
                "Error: transmission could not be set (returned status %s)."
                % foil_status
            )



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Oxford Cryostream, controlled by bliss.
Example xml_ configuration:

.. code-block:: xml

 <object class="ESRF.OxfordCryostream">
   <username>Cryostream</username>
   <object role="controller" href="/bliss"/>
   <cryostat>cryostream</cryostat>
   <interval>120</interval>
   <object role="monitor_temperature" href="/monitor_temperature"/>
 </object>
"""

import logging
import sys

from gevent import (
    Timeout,
    sleep,
    spawn,
)

from mxcubecore.HardwareObjects.abstract.AbstractActuator import AbstractActuator

CRYO_STATUS = ["OFF", "SATURATED", "READY", "WARNING", "FROZEN", "UNKNOWN"]
PHASE_ACTION = {
    "RAMP": "ramp",
    "COOL": "cool",
    "HOLD": "hold",
    "PLAT": "plat",
    "PURGE": "purge",
    "END": "end",
}


class OxfordCryostream(AbstractActuator):
    """Control of the Oxford Cryostream model 700, 800 and 1000"""

    def __init__(self, name):
        super().__init__(name)

        self.temp = None
        self.temp_threshold = None
        self._monitor_obj = None
        self._timeout = 5  # [s]
        self._hw_ctrl = None
        self.ctrl = None
        self.interval = None

    def _do_polling(self):
        """Do endless polling with predefined interval"""
        while True:
            try:
                self.force_emit_signals()
            except Exception:
                sys.excepthook(*sys.exc_info())
            sleep(self.interval)

    def init(self):
        """Initialisation"""
        controller = self.get_object_by_role("controller")
        cryostat = self.get_property("cryostat")
        self.interval = self.get_property("interval", 10)
        try:
            self.ctrl = getattr(controller, cryostat)
            spawn(self._do_polling)
            self._hw_ctrl = self.ctrl.controller._hw_controller
        except AttributeError as err:
            raise RuntimeError("Cannot use cryostream") from err

        self._monitor_obj = self.get_object_by_role("monitor_temperature")
        self.temp_threshold = self.get_property("temperature_threshold", 0.0)

    def force_emit_signals(self):
        """Forces to emit all signals."""
        self.emit("valueChanged", (self.get_value(),))
        self.emit("stateChanged", (self.get_state(),))

    def get_temperature(self):
        """Read the temperature.
        Returns:
            (float): The temperature [deg K]
        """
        try:
            return self.ctrl.input.read()
        except Exception:
            # try to read again
            temp = self.ctrl.input.read()
            if temp is None:
                return 9999.0
        return temp

    def get_value(self):
        return self.get_temperature()

    def _set_value(self, value=None):
        """Define the setpoint.
        Args:
           value(float): target temperature [deg K]
        """
        if value is not None:
            self.ctrl.setpoint = value

    def rampstate(self):
        """Read the state of the ramping.
        Returns:
            (str): Ramping state.
        """
        return self.ctrl.is_ramping()

    def start_action(self, phase="RAMP", target=None, rate=None):
        """Run phase action action.
        Args:
            phase(str): The phase action. Default value - RAMP
            target(float): Target temperature.
            rate:(float): Ramp rate.
        """
        if phase in PHASE_ACTION:
            action = getattr(self._hw_ctrl, PHASE_ACTION[phase])
            if rate:
                action(target, rate=rate)
            elif target:
                action(target)
            else:
                action()

    def stop_action(self, phase="HOLD"):
        """Stop action.
        Args:
            phase(str): Phase action.
        """
        if phase in PHASE_ACTION:
            getattr(self._hw_ctrl, PHASE_ACTION[phase])

    def pause(self, execute=True):
        """Pause the ramping.
        Args:
            execute(bool): True to pause, False to resume.
        """
        if execute:
            self._hw_ctrl.pause()
        else:
            self._hw_ctrl.resume()

    def get_specific_state(self):
        """Read the state of the controller.
        Returns:
            (str): The state.
        """
        try:
            return self._hw_ctrl.read_run_mode().upper()
        except (AttributeError, TypeError):
            return "UNKNOWN"

    def get_static_parameters(self):
        """Get predefined parameters.
        Returns:
            {list): Predefimed parameters.
        """
        return ["oxford", "K", "hour"]

    def get_params(self):
        """Read from the controller.
        Returns:
            (list): [target_temperature, ramp_rate, phase, run_mode]
        """
        target_temperature = self.ctrl.setpoint
        ramp_rate = self.ctrl.ramprate
        phase = self._hw_ctrl.read_phase().upper()
        run_mode = self._hw_ctrl.read_run_mode()
        self.temp = self.ctrl.input.read()
        return [target_temperature, ramp_rate, phase, run_mode]

    def check_temperature(self, threshold=None):
        """Check if the temperature is under the threshold.
        Args:
            threshold (float): Temperature threshold (optional)
        Returns:
            (bool): True if under the threshold, False otherwise.
        """
        threshold = threshold or self.temp_threshold
        logging.getLogger("user_level_log").info("Cryo temperature reading ...")
        cryo_temp = self.get_value()
        if cryo_temp > threshold:
            logging.getLogger("user_level_log").info("Cryo temperature too high ...")
            return False
        return True

    def wait_temperature(self, threshold=None, timeout=None):
        """Wait until the temperature is under the threshold.
        Args:
            threshold (float): Temperature threshold (optional)
            timeout (float): optional - timeout [s],
                             If timeout == 0: return at once and do not wait
                                              (default);
                             if timeout is None: wait forever.
        """
        if self._monitor_obj:
            try:
                check = self._monitor_obj.get_value().value
            except AttributeError:
                check = False
            if check is True:
                threshold = threshold or self.temp_threshold
                timeout = timeout or self._timeout
                with Timeout(timeout, RuntimeError("Temperature timeout")):
                    while not self.check_temperature(threshold=threshold):
                        sleep(0.5)



import logging
import os
import subprocess
import time

import gevent
from PyTango import DeviceProxy

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.CommandContainer import ConnectionError
from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector
from mxcubecore.TaskUtils import task


class LimaPilatusDetector(AbstractDetector):
    def __init__(self, name):
        AbstractDetector.__init__(self, name)
        self.header = {}
        self.start_angles = []

    def init(self):
        AbstractDetector.init(self)

        lima_device = self.get_property("lima_device")
        pilatus_device = self.get_property("pilatus_device")

        if None in (lima_device, pilatus_device):
            return

        try:
            for channel_name in (
                "latency_time",
                "State",
                "acq_status",
                "acq_trigger_mode",
                "saving_mode",
                "acq_nb_frames",
                "acq_expo_time",
                "saving_directory",
                "saving_prefix",
                "saving_suffix",
                "saving_next_number",
                "saving_index_format",
                "saving_format",
                "saving_overwrite_policy",
                "saving_header_delimiter",
                "last_image_saved",
                "image_roi",
            ):
                self.add_channel(
                    {"type": "tango", "name": channel_name, "tangoname": lima_device},
                    channel_name,
                )

            for channel_name in ("fill_mode", "threshold"):
                self.add_channel(
                    {
                        "type": "tango",
                        "name": channel_name,
                        "tangoname": pilatus_device,
                    },
                    channel_name,
                )

            pilatus_tg_device = DeviceProxy(pilatus_device)

            if hasattr(pilatus_tg_device, "working_energy"):
                self.add_channel(
                    {
                        "type": "tango",
                        "name": "working_energy",
                        "tangoname": pilatus_device,
                    },
                    "working_energy",
                )
            self.add_channel(
                {
                    "type": "tango",
                    "name": "energy_threshold",
                    "tangoname": pilatus_device,
                },
                "energy_threshold",
            )

            self.add_command(
                {"type": "tango", "name": "prepare_acq", "tangoname": lima_device},
                "prepareAcq",
            )
            self.add_command(
                {"type": "tango", "name": "start_acq", "tangoname": lima_device},
                "startAcq",
            )
            self.add_command(
                {"type": "tango", "name": "stop_acq", "tangoname": lima_device},
                "stopAcq",
            )
            self.add_command(
                {"type": "tango", "name": "reset", "tangoname": lima_device}, "reset"
            )
            self.add_command(
                {"type": "tango", "name": "set_image_header", "tangoname": lima_device},
                "SetImageHeader",
            )

            self.get_channel_object("image_roi").connect_signal(
                "update", self.roi_mode_changed
            )

            self.get_command_object("prepare_acq").set_device_timeout(10000)
            self._emit_status()

        except ConnectionError:
            self.update_state(HardwareObjectState.FAULT)
            logging.getLogger("HWR").error(
                "Could not connect to detector %s" % lima_device
            )
            self._emit_status()

    def has_shutterless(self):
        return True

    def wait_ready(self, timeout=3500):
        with gevent.Timeout(timeout, RuntimeError("Detector not ready")):
            while self.get_channel_value("acq_status") != "Ready":
                time.sleep(1)

    def last_image_saved(self):
        try:
            img = self.get_channel_object("last_image_saved").get_value() + 1
            return img
        except Exception:
            return 0

    def get_deadtime(self):
        return float(self.get_property("deadtime"))

    def roi_mode_changed(self, mode):
        """ROI mode change event"""
        self.roi_mode = self.roi_modes_list.index(str(mode))
        self.emit("detectorRoiModeChanged", (self.roi_mode,))

    def prepare_acquisition(
        self,
        take_dark,
        start,
        osc_range,
        exptime,
        npass,
        number_of_images,
        comment,
        mesh,
        mesh_num_lines,
    ):
        if mesh:
            trigger_mode = self.get_property("mesh_trigger_mode", "EXTERNAL_GATE")
        else:
            trigger_mode = self.get_property("osc_trigger_mode", "EXTERNAL_GATE")

        diffractometer_positions = HWR.beamline.diffractometer.get_positions()
        self.start_angles = []
        for i in range(number_of_images):
            self.start_angles.append("%0.4f deg." % (start + osc_range * i))
        self.header = {}
        self.header["file_comments"] = comment
        self.header["N_oscillations"] = number_of_images
        self.header["Oscillation_axis"] = "omega"
        self.header["Chi"] = "0.0000 deg."
        kappa_phi = diffractometer_positions.get("kappa_phi", -9999)
        if kappa_phi is None:
            kappa_phi = -9999
        kappa = diffractometer_positions.get("kappa", -9999)
        if kappa is None:
            kappa = -9999
        self.header["Phi"] = "%0.4f deg." % kappa_phi
        self.header["Kappa"] = "%0.4f deg." % kappa
        self.header["Alpha"] = "0.0000 deg."
        self.header["Polarization"] = HWR.beamline.collect.bl_config.polarisation
        self.header["Detector_2theta"] = "0.0000 deg."
        self.header["Angle_increment"] = "%0.4f deg." % osc_range
        self.header["Transmission"] = HWR.beamline.transmission.get_value()

        self.header["Flux"] = HWR.beamline.flux.get_value()
        self.header["Beam_xy"] = "(%.2f, %.2f) pixels" % self.get_beam_position()
        self.header["Detector_Voffset"] = "0.0000 m"
        self.header["Energy_range"] = "(0, 0) eV"
        self.header["Detector_distance"] = "%f m" % (self.distance.get_value() / 1000.0)
        self.header["Wavelength"] = "%f A" % HWR.beamline.energy.get_wavelength()
        self.header["Trim_directory:"] = "(nil)"
        self.header["Flat_field:"] = "(nil)"
        self.header["Excluded_pixels:"] = " badpix_mask.tif"
        self.header["N_excluded_pixels:"] = "= 321"
        self.header["Threshold_setting"] = "%d eV" % self.get_channel_value("threshold")
        self.header["Count_cutoff"] = "1048500"
        self.header["Tau"] = "= 0 s"
        self.header["Exposure_period"] = "%f s" % (exptime + self.get_deadtime())
        self.header["Exposure_time"] = "%f s" % exptime

        self.reset()
        self.wait_ready()

        self.set_energy_threshold(HWR.beamline.energy.get_value())

        self.set_channel_value("acq_trigger_mode", trigger_mode)

        if self.get_property("set_latency_time", False):
            self.set_channel_value("latency_time", self.get_deadtime())

        self.set_channel_value("saving_mode", "AUTO_FRAME")
        self.set_channel_value("acq_nb_frames", number_of_images)
        self.set_channel_value("acq_expo_time", exptime)
        self.set_channel_value("saving_overwrite_policy", "OVERWRITE")

    def set_energy_threshold(self, energy):
        """Set the energy threshold.
        Args:
            energy (int): Energy [eV] or [keV]
        """
        minE = self.get_property("minE")
        # some versions of Lima Pilatus server take the energy ergument in keV
        # some in eV. From minE we can set a conversion factor.
        factor = 1000 if minE > 100 else 1.0

        energy_threshold = self.get_channel_value("energy_threshold")

        # check if need to convert energy in eV.
        if energy < 100:
            energy *= factor

        if energy < minE:
            energy = minE

        if abs(energy_threshold - energy) > 0.1:
            self.set_channel_value("energy_threshold", energy)
            while abs(self.get_channel_value("energy_threshold") - energy) > 0.1:
                time.sleep(1)

        self.set_channel_value("fill_mode", "ON")

    @task
    def set_detector_filenames(self, frame_number, start, filename):
        prefix, suffix = os.path.splitext(os.path.basename(filename))
        prefix = "_".join(prefix.split("_")[:-1]) + "_"
        dirname = os.path.dirname(filename)
        if dirname.startswith(os.path.sep):
            dirname = dirname[len(os.path.sep) :]

        saving_directory = os.path.join(self.get_property("buffer", "/"), dirname)

        subprocess.Popen(
            "ssh %s@%s mkdir --parents %s"
            % (os.environ["USER"], self.get_property("control"), saving_directory),
            shell=True,
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
        ).wait()

        self.set_channel_value("saving_directory", saving_directory)
        self.set_channel_value("saving_prefix", prefix)
        self.set_channel_value("saving_suffix", suffix)
        self.set_channel_value("saving_next_number", frame_number)
        self.set_channel_value("saving_index_format", "%04d")
        self.set_channel_value("saving_format", "CBF")
        self.set_channel_value("saving_header_delimiter", ["|", ";", ":"])

        headers = []

        for i, start_angle in enumerate(self.start_angles):
            header = "\n%s\n" % self.get_property("serial")
            header += "# %s\n" % time.strftime("%Y/%b/%d %T")
            header += "\n%s\n" % self.get_property("sensor")
            header += "\n%s\n" % self.get_property("pixel_size")
            self.header["Start_angle"] = start_angle

            for key, value in self.header.items():
                header += "# %s %s\n" % (key, value)

            headers.append("%d : array_data/header_contents|%s;" % (i, header))

        self.header = headers

    def start_acquisition(self):
        self.wait_ready()
        self.execute_command("stop_acq")
        self.execute_command("prepare_acq")
        self.execute_command("set_image_header", self.header)
        self.execute_command("start_acq")
        self._emit_status()

    def stop_acquisition(self):
        try:
            self.execute_command("stop_acq")
        except Exception:
            pass

        time.sleep(1)
        self.execute_command("reset")

        self.wait_ready()

        self._emit_status()

    def reset(self):
        self.stop_acquisition()

    @property
    def status(self):
        try:
            acq_status = self.get_channel_value("acq_status")
        except Exception:
            acq_status = "OFFLINE"

        status = {
            "acq_satus": acq_status.upper(),
        }

        return status

    def _emit_status(self):
        self.emit("statusChanged", self.status)

    def recover_from_failure(self):
        self.prepare_acquisition(
            False,
            0,
            0,
            0.5,
            None,
            1,
            "",
            HWR.beamline.energy.get_value(),
            "INTERNAL_TRIGGER",
        )
        self.start_acquisition()
        self.wait_ready()
        self.stop_acquisition()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

# import time
# import logging
# import weakref

# try:
#     import Queue as queue
# except ImportError:
#     import queue


from mxcubecore.CommandContainer import (
    ChannelObject,
    CommandObject,
)

__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"


class MockupCommand(CommandObject):
    def __init__(self, name, command_name, list_args=None, timeout=1000, **kwargs):
        CommandObject.__init__(self, name, **kwargs)

        self.command_name = command_name
        self.result = None

    def __call__(self, *args, **kwargs):
        self.result = args[0]

    def get(self):
        return self.result

    def abort(self):
        pass

    def is_connected(self):
        return True


class MockupChannel(ChannelObject):
    def __init__(self, name, username=None, timeout=1000, **kwargs):
        ChannelObject.__init__(self, name, username, **kwargs)

        self.timeout = int(timeout)
        self.value = kwargs["default_value"]

    def get_value(self, force=False):
        return self.value

    def set_value(self, new_value):
        self.value = new_value
        self.emit("update", self.value)

    def is_connected(self):
        # FIXME linkid is not defined for this class but only for Tine
        return self.linkid > 0



import logging

from mxcubecore.BaseHardwareObjects import HardwareObject


class SOLEILSafetyShutter(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.pss = None
        self.shutter = None

    def init(self):
        try:
            self.shutter = self.get_object_by_role("shutter")
            self.pss = self.get_object_by_role("pss")
            logging.debug("shutter is " + str(self.shutter))
            logging.debug("pss is " + str(self.pss))
            self.connect(self.shutter, "shutterStateChanged", self.shutterStateChanged)
            self.connect(self.pss, "wagoStateChanged", self.shutterStateChanged)
        except Exception:
            import traceback

            print(traceback.print_exc())
            logging.debug(traceback.format_exc())
            logging.getLogger().warning("pss device not configured")

    def getShutterState(self):
        logging.debug(" shutter is %s " % str(self.shutter))

        if self.pss.getWagoState() != "ready":
            return "disabled"

        if self.shutter is None:
            return "unknown"
        return self.shutter.getShutterState()

    def shutterStateChanged(self, value):
        if self.shutter is None:
            return
        #
        # emit signal
        #
        self.shutterStateValue = value  # str(value)
        self.emit("shutterStateChanged", (self.getShutterState(),))
        self.emit("stateChanged", (self.getShutterState(),))

    def openShutter(self):
        if self.shutter is None:
            return
        if self.pss is None:
            logging.error("no pss device for safety shutter. check configuration")
            return

        if self.pss.getWagoState() == "ready":
            self.shutter.openShutter()
            logging.info("Opening shutter ok")
        else:
            logging.warning("cannot open safety shutter. Check interlock")

    def closeShutter(self):
        if self.shutter is None:
            return
        self.shutter.closeShutter()



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
from copy import deepcopy
from datetime import datetime

import gevent
import numpy
import tine
from scipy.interpolate import interp1d

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractFlux import AbstractFlux

__credits__ = ["EMBL Hamburg"]
__category__ = "General"

diode_calibration_amp_per_watt = interp1d(
    [4.0, 6.0, 8.0, 10.0, 12.0, 12.5, 15.0, 16.0, 20.0, 30.0],
    [0.2267, 0.2116, 0.1405, 0.086, 0.0484, 0.0469, 0.0289, 0.0240, 0.01248, 0.00388],
)

diode_calibration_amp_per_watt = interp1d(
    [4.0, 6.0, 8.0, 10.0, 12.0, 12.5, 15.0, 16.0, 20.0, 30.0],
    [0.2267, 0.2116, 0.1405, 0.086, 0.0484, 0.0469, 0.0289, 0.0240, 0.01248, 0.00388],
)

air_absorption_coeff_per_meter = interp1d(
    [
        4.0,
        4.2,
        4.4,
        4.6,
        4.8,
        5.0,
        5.4,
        5.8,
        6.2,
        6.6,
        7.0,
        8.0,
        9.2,
        11.8,
        14.4,
        17.0,
        19.6,
        22.2,
        24.8,
        27.4,
        30,
    ],
    [
        9.19440446,
        7.94983601,
        6.91804807,
        6.10374226,
        5.32906528,
        4.71308953,
        3.73630655,
        3.00942560,
        2.45767288,
        2.0317802,
        1.69805057,
        1.12911273,
        0.73628084,
        0.34554261,
        0.19176669,
        0.12030697,
        0.08331135,
        0.06203213,
        0.04926173,
        0.04114024,
        0.0357374,
    ],
)

carbon_window_transmission = interp1d(
    [4.0, 6.6, 9.2, 11.8, 14.4, 17.0, 19.6, 22.2, 24.8, 27.4, 30],
    [
        0.74141,
        0.93863,
        0.97775,
        0.98946,
        0.99396,
        0.99599,
        0.99701,
        0.99759,
        0.99793,
        0.99815,
        0.99828,
    ],
)

# # Replaced by AbstractFLux.get_dose_rate_per_photon_per_mmsq (same values0
#
# dose_rate_per_10to14_ph_per_mmsq = interp1d(
#     [4.0, 6.6, 9.2, 11.8, 14.4, 17.0, 19.6, 22.2, 24.8, 27.4, 30.0],
#     [
#         459000.0,
#         162000.0,
#         79000.0,
#         45700.0,
#         29300.0,
#         20200.0,
#         14600.0,
#         11100.0,
#         8610.0,
#         6870.0,
#         5520.0,
#     ],
# )


class EMBLFlux(AbstractFlux):
    def __init__(self, name):
        AbstractFlux.__init__(self, name)

        self.measured_flux_dict = None
        self.measured_flux_list = None
        self.current_flux_dict = None

        self.flux_value = 0
        # self.ampl_chan_index = None
        # self.intensity_ranges = []
        self.intensity_value = None

        self.flux_record_status = None

        self.origin_flux_value = None
        self.origin_beam_info = None
        self.origin_transmission = None
        self.measuring = False
        self.transmission_value = None

        self.chan_intens_range = None
        self.chan_intens_mean = None
        self.cmd_set_intens_acq_time = None
        self.cmd_set_intens_range = None
        self.cmd_set_intens_resolution = None

    def init(self):
        """Reads config xml, initiates all necessary hwobj, channels and cmds"""
        super(EMBLFlux, self).init()

        """
        self.intensity_ranges = []
        try:
            for intens_range in self["intensity"]["ranges"]:
                temp_intens_range = {}
                temp_intens_range["max"] = intens_range.CurMax
                temp_intens_range["index"] = intens_range.CurIndex
                temp_intens_range["offset"] = intens_range.CurOffset
                self.intensity_ranges.append(temp_intens_range)
            self.intensity_ranges = sorted(
                self.intensity_ranges, key=lambda item: item["max"]
            )
        except Exception:
            logging.getLogger("HWR").error("BeamlineTest: No intensity ranges defined")
        """

        self.chan_intens_mean = self.get_channel_object("intensMean")
        self.chan_intens_mean.connect_signal("update", self.intens_mean_changed)

        self.chan_intens_range = self.get_channel_object("intensRange")
        self.chan_flux_transmission = self.get_channel_object("fluxTransmission")

        self.cmd_set_intens_resolution = self.get_command_object("setIntensResolution")
        self.cmd_set_intens_acq_time = self.get_command_object("setIntensAcqTime")
        self.cmd_set_intens_range = self.get_command_object("setIntensRange")
        self.cmd_flux_record = self.get_command_object("fluxRecord")

        self.connect(
            HWR.beamline.transmission, "valueChanged", self.transmission_changed
        )

        # self.init_flux_values()

        self.chan_flux_status = self.get_channel_object("fluxStatus")
        self.chan_flux_status.connect_signal("update", self.flux_status_changed)

        self.chan_flux_message = self.get_channel_object("fluxMessage")
        self.chan_flux_message.connect_signal("update", self.flux_message_changed)

        self.connect(HWR.beamline.beam, "beamInfoChanged", self.beam_info_changed)

        self.connect(
            HWR.beamline.beam.aperture,
            "diameterIndexChanged",
            self.aperture_diameter_changed,
        )

        if HWR.beamline.beam.definer is not None:
            self.connect(
                HWR.beamline.beam.definer,
                "focusingModeChanged",
                self.focusing_mode_changed,
            )

        self.init_flux_values()

    def init_flux_values(self):
        if not self.chan_flux_status.get_value():
            logging.getLogger("GUI").error(
                "No valid flux value available. Please remeasure flux!"
            )
            return
        flux_values = self.cmd_flux_record.get()
        flux_transmission = self.chan_flux_transmission.get_value()
        aperture_diameter_list = HWR.beamline.beam.aperture.get_diameter_size_list()

        self.measured_flux_list = []
        for index, flux_value in enumerate(flux_values):
            self.measured_flux_list.append(
                {
                    "flux": flux_value,
                    "transmission": flux_transmission,
                    "size_x": aperture_diameter_list[index] / 1000.0,
                    "size_y": aperture_diameter_list[index] / 1000.0,
                }
            )

    def flux_message_changed(self, message):
        if message is not "":
            logging.getLogger("GUI").error("Flux-record message: %s" % message)

    def flux_status_changed(self, status):
        if not status and self.flux_record_status:
            logging.getLogger("GUI").error(
                "Flux value invalidated. Please remeasure flux!"
            )
            self.reset_flux()
        self.flux_record_status = status

    def aperture_diameter_changed(self, index, size):
        """Updates flux if the aperture diameter has been changed"""
        if self.measured_flux_list and not self.measuring:
            if len(self.measured_flux_list) > 1:
                self.measured_flux_dict = self.measured_flux_list[index]
                self.update_flux_value()

    def beam_info_changed(self, beam_info):
        """Updates flux value if the beam size changes"""
        self.beam_info = beam_info
        self.update_flux_value()

    def transmission_changed(self, transmission):
        """Updates flux value if the transmission has been changed"""
        self.transmission_value = transmission
        self.update_flux_value()

    def intens_mean_changed(self, value):
        pass

    def focusing_mode_changed(self, mode, size):
        logging.getLogger("GUI").warning(
            "Beamline focus mode changed. Please remeasure flux!"
        )
        self.reset_flux()

    def reset_flux(self):
        self.current_flux_dict = None
        self.measured_flux_dict = None
        self.measured_flux_list = []
        self.emit(
            "fluxInfoChanged",
            {"measured": self.measured_flux_dict, "current": self.current_flux_dict},
        )

    def get_value(self):
        """Returns flux value as float"""
        if self.current_flux_dict is not None:
            return self.current_flux_dict["flux"]
        else:
            return 1

    def update_flux_value(self):
        if self.measured_flux_dict is not None and self.transmission_value is not None:
            self.current_flux_dict = deepcopy(self.measured_flux_dict)
            if int(self.transmission_value) != int(
                self.measured_flux_dict["transmission"]
            ):
                self.current_flux_dict["flux"] = (
                    self.measured_flux_dict["flux"]
                    * self.transmission_value
                    / self.measured_flux_dict["transmission"]
                )
                self.current_flux_dict["transmission"] = self.transmission_value

            if len(self.measured_flux_list) == 1:
                origin_area = (
                    self.measured_flux_dict["size_x"]
                    * self.measured_flux_dict["size_y"]
                )
                current_area = self.beam_info["size_x"] * self.beam_info["size_y"]
                if origin_area != current_area:
                    self.current_flux_dict["size_x"] = self.beam_info["size_x"]
                    self.current_flux_dict["size_y"] = self.beam_info["size_y"]
                    self.current_flux_dict["flux"] = (
                        self.measured_flux_dict["flux"] * current_area / origin_area
                    )
            self.emit(
                "fluxInfoChanged",
                {
                    "measured": self.measured_flux_dict,
                    "current": self.current_flux_dict,
                },
            )

    def measure_flux(self, wait=True):
        gevent.spawn(self.measure_flux_task, wait)

    def measure_flux_task(self, wait=True):
        if not HWR.beamline.safety_shutter.is_opened():
            msg = "Unable to measure flux! Safety shutter is closed."
            self.print_log("GUI", "error", msg)
            return

        if not HWR.beamline.detector.is_cover_closed():
            msg = "Unable to measure flux! Detecor cover is open."
            self.print_log("GUI", "error", "Unable to measure flux!")
            self.print_log("GUI", "error", msg)
            return

        if HWR.beamline.session.beamline_name == "P14":
            if HWR.beamline.detector.distance.get_value() > 501:
                self.print_log(
                    "GUI",
                    "error",
                    "Detector is too far away for flux measurements. Move to 500 mm or closer.",
                )
                return

        self.measuring = True
        intens_value = 0
        max_frame_rate = 1 / HWR.beamline.detector.get_exposure_time_limits()[0]

        current_phase = HWR.beamline.diffractometer.current_phase
        current_transmission = HWR.beamline.transmission.get_value()
        current_aperture_index = HWR.beamline.beam.aperture.get_diameter_index()

        self.emit("progressInit", "Measuring flux. Please wait...", 10, True)

        # Set transmission to 100%
        # -----------------------------------------------------------------
        self.emit("progressStep", 1, "Setting transmission to 100%")

        HWR.beamline.transmission.set_value(100, timeout=20)

        # Close the fast shutter
        # -----------------------------------------------------------------
        HWR.beamline.fast_shutter.close(wait=True)
        logging.getLogger("HWR").debug("Measure flux: Fast shutter closed")
        gevent.sleep(0.2)
        HWR.beamline.diffractometer.wait_ready(10)

        # Move back light in, check beamstop position
        # -----------------------------------------------------------------
        logging.getLogger("HWR").info("Measure flux: Moving backlight out...")
        self.emit("progressStep", 1, "Moving backlight out")
        HWR.beamline.back_light.move_in()
        logging.getLogger("HWR").debug("Measure flux: Backlight moved out")

        beamstop_position = HWR.beamline.beamstop.get_value()
        if beamstop_position == "BEAM":
            self.emit("progressStep", 2, "Moving beamstop OFF")
            HWR.beamline.beamstop.set_position("OFF")
            HWR.beamline.diffractometer.wait_ready(30)
            logging.getLogger("HWR").info("Measure flux: Beamstop moved off")

        # Check scintillator position
        # -----------------------------------------------------------------
        scintillator_position = HWR.beamline.diffractometer.get_scintillator_position()
        if scintillator_position == "SCINTILLATOR":
            self.emit("progressStep", 3, "Setting the photodiode")
            HWR.beamline.diffractometer.set_scintillator_position("PHOTODIODE")
            gevent.sleep(1)
            HWR.beamline.diffractometer.wait_ready(30)
            logging.getLogger("HWR").debug(
                "Measure flux: Scintillator set to photodiode"
            )

        self.measured_flux_list = []

        # -----------------------------------------------------------------
        if HWR.beamline.session.beamline_name == "P13":
            HWR.beamline.beam.aperture.set_in()
            HWR.beamline.diffractometer.wait_ready(30)
            HWR.beamline.beam.aperture.set_diameter_index(0)
            HWR.beamline.fast_shutter.openShutter(wait=True)

            for index, diameter_size in enumerate(
                HWR.beamline.beam.aperture.get_diameter_list()
            ):
                # 5. open the fast shutter -----------------------------------------
                self.emit(
                    "progressStep",
                    4 + index,
                    "Measuring flux with %d micron aperture" % diameter_size,
                )
                HWR.beamline.beam.aperture.set_diameter_index(index)
                HWR.beamline.diffractometer.wait_ready(10)

                gevent.sleep(1)
                intens_value = self.chan_intens_mean.get_value(force=True)
                logging.getLogger("HWR").info("Measured current: %s" % intens_value)
                # HWR.beamline.fast_shutter.closeShutter(wait=True)
                intensity_value = intens_value[0] + 1.860e-5  # 2.780e-6
                self.measured_flux_list.append(self.get_flux_result(intensity_value))
                gevent.sleep(1)
            HWR.beamline.fast_shutter.closeShutter(wait=True)

            try:
                self.cmd_flux_record([_x["flux"] for _x in self.measured_flux_list])
                gevent.sleep(2)
            except Exception:
                pass

            max_frame_rate = 25
        else:
            self.emit("progressStep", 5, "Measuring the intensity")
            current_aperture_index = 0
            HWR.beamline.fast_shutter.open(wait=True)
            logging.getLogger("HWR").debug("Measure flux: Fast shutter opened")

            gevent.sleep(0.5)
            intens_value = self.chan_intens_mean.get_value()

            intens_range_now = self.chan_intens_range.get_value()
            HWR.beamline.fast_shutter.close(wait=True)
            logging.getLogger("HWR").debug("Measure flux: Fast shutter closed")

            intensity_value = intens_value[0] + 2.780e-6
            self.measured_flux_list.append(self.get_flux_result(intensity_value))

        self.emit("progressStep", 10, "Restoring original state")
        self.print_log("GUI", "info", "Flux measurement results:")
        self.print_log(
            "GUI",
            "info",
            "Beam size | Flux (ph/s) | "
            + "Dose rate (KGy/s) | Time to reach 20 MGy (s) | "
            + "Number of frames @ %d Hz" % max_frame_rate,
        )

        for index, item in enumerate(self.measured_flux_list):
            msg = "  * %d x %d | %1.1e  | %1.1e  | %.1f  | %d" % (
                item["size_x"] * 1000,
                item["size_y"] * 1000,
                item["flux"],
                item["dose_rate"],
                item["time_to_reach_limit"],
                item["frames_to_reach_limit"],
            )

            if index > 0:
                # low_value = item["flux"] < 1e9
                # low_value = item["intensity"] - 1.860e-5 < 1e-6
                low_value = item["intensity"] < 1e-6
                out_of_range = False

                if (
                    self.measured_flux_list[0]["flux"]
                    <= self.measured_flux_list[-1]["flux"]
                    or self.measured_flux_list[index - 1]["flux"]
                    <= self.measured_flux_list[index]["flux"]
                ):
                    out_of_range = True
                if low_value or out_of_range:
                    msg += " (intensity: %1.1e)" % item["intensity"]
                    self.print_log("GUI", "error", msg)
                else:
                    self.print_log("GUI", "info", msg)
            else:
                self.print_log("GUI", "info", msg)

        self.measured_flux_dict = self.measured_flux_list[current_aperture_index]
        self.current_flux_dict = self.measured_flux_list[current_aperture_index]

        self.emit(
            "fluxInfoChanged",
            {"measured": self.measured_flux_dict, "current": self.current_flux_dict},
        )
        self.measuring = False

        # 7 Restoring previous states ----------------------------------------
        HWR.beamline.transmission.set_value(current_transmission)
        HWR.beamline.diffractometer.set_phase(current_phase)
        HWR.beamline.diffractometer.wait_ready(10)
        if HWR.beamline.session.beamline_name == "P13":
            HWR.beamline.beam.aperture.set_diameter_index(current_aperture_index)
        self.emit("progressStop", ())

    def get_flux_result(self, intensity_value):
        energy = HWR.beamline.energy.get_value()
        detector_distance = HWR.beamline.detector.distance.get_value()
        beam_size = HWR.beamline.beam.get_beam_size()
        transmission = HWR.beamline.transmission.get_value()

        air_trsm = numpy.exp(
            -air_absorption_coeff_per_meter(energy) * detector_distance / 1000.0
        )
        carb_trsm = carbon_window_transmission(energy)
        flux = (
            0.624151
            * 1e16
            * intensity_value
            / diode_calibration_amp_per_watt(energy)
            / energy
            / air_trsm
            / carb_trsm
        )

        flux = flux * 1.8
        dose_rate = (
            1e-3
            # * 1e-14
            * self.get_dose_rate_per_photon_per_mmsq(energy)
            * flux
            / beam_size[0]
            / beam_size[1]
        )
        max_frame_rate = 1 / HWR.beamline.detector.get_exposure_time_limits()[0]

        result = {
            "energy": energy,
            "detector_distance": detector_distance,
            "size_x": beam_size[0],
            "size_y": beam_size[1],
            "transmission": transmission,
            "intensity": intensity_value,
            "flux": flux,
            "dose_rate": dose_rate,
            "time_to_reach_limit": 20000.0 / dose_rate,
            "frames_to_reach_limit": int(max_frame_rate * 20000.0 / dose_rate),
            "max_frame_rate": max_frame_rate,
        }

        return result



from __future__ import absolute_import

import functools
import logging
import os
import sys
from logging.handlers import RotatingFileHandler

from colorama import (
    Back,
    Fore,
    Style,
)

from mxcubecore import BaseHardwareObjects as BHWO
from mxcubecore import HardwareRepository as HWR
from mxcubecore import __version__

__version__ = __version__.__version__

hwrpath = os.path.dirname(__file__)
sys.path.insert(0, hwrpath)

#
# this makes it possible for Hardware Objects to import
# standard Hardware Objects easily
#


class ColorFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: "\033[1m" + "%s" + Style.RESET_ALL,
        logging.INFO: Fore.BLUE + "%s" + Style.RESET_ALL,
        logging.WARNING: Fore.YELLOW + "%s" + Style.RESET_ALL,
        logging.ERROR: "\033[1m" + Fore.RED + "%s" + Style.RESET_ALL,
        logging.CRITICAL: "\033[1m" + Fore.RED + "%s" + Style.RESET_ALL,
    }

    def format(self, record):
        formatter = logging.Formatter(self.FORMATS.get(record.levelno) % self._fmt)
        return formatter.format(record)


def trace_call_log(_func=None, *, level: int = logging.DEBUG):
    def decorator_wrapper(func):
        @functools.wraps(func)
        def fun_wrapper(*args, **kwargs):
            if len(args) == 0 or (not isinstance(args[0], BHWO.HardwareObject)):
                err_msg = (
                    "Not valid method. Decorator can be applied to an"
                    " HardwareObject instance's methods only"
                )
                raise TypeError(err_msg)
            args = list(args)
            self = args.pop(0)
            hwo_name = f" ({self.name.strip('/')})" if self.name else ""
            # Remove named parameters which are not in the signature of the method
            code_obj = func.__code__
            kwargs = {
                k: v
                for k, v in kwargs.items()
                if k in code_obj.co_varnames[: code_obj.co_argcount]
            }

            args_str = ",".join([str(arg) for arg in args])
            kwargs_str = ",".join(["%s=%s" % (k, v) for k, v in kwargs.items()])
            params_str = f"{args_str}{', ' if args and kwargs else ''}{kwargs_str}"
            method_name = func.__name__
            signature = f"{method_name}({params_str})"
            self.log.log(level, f"{hwo_name} In {signature}")
            return func(self, *args, **kwargs)

        return fun_wrapper

    if _func is None:
        return decorator_wrapper
    return decorator_wrapper(_func)


def getStdHardwareObjectsPath():
    import HardwareObjects  # first looks in containing package

    return os.path.dirname(HardwareObjects.__file__)


sys.path.insert(0, getStdHardwareObjectsPath())

hwobj_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "HardwareObjects")
hwobj_dir_list = [hwobj_dir]
for subdir in ("sample_changer", "mockup"):
    hwobj_dir_list.append(os.path.join(hwobj_dir, subdir))
HWR.add_hardware_objects_dirs(hwobj_dir_list)

#
# create the HardwareRepository logger
#
_hwr_logger = logging.getLogger("HWR")
_hwr_logger.setLevel(logging.DEBUG)
_oldLevel = logging.DEBUG
_hwr_formatter = logging.Formatter("%(asctime)s |%(levelname)-7s| %(message)s")

if len(logging.root.handlers) == 0:
    #
    # log to stdout
    #
    _hdlr = logging.StreamHandler(sys.stdout)
    _hdlr.setFormatter(ColorFormatter("%(asctime)s |%(levelname)-7s| %(message)s"))
    _hwr_logger.addHandler(_hdlr)


def removeLoggingHandlers():
    for handler in _hwr_logger.handlers:
        _hwr_logger.removeHandler(handler)


def setLoggingOff():
    global _oldLevel
    _oldLevel = _hwr_logger.getEffectiveLevel()
    _hwr_logger.setLevel(
        1000
    )  # disable all logging events less severe than 1000 (CRITICAL is 50...)


def setLoggingOn():
    _hwr_logger.setLevel(_oldLevel)


def addLoggingHandler(handler):
    _hwr_logger.addHandler(handler)


def setLoggingHandler(handler):
    global _hdlr

    removeLoggingHandlers()  # _logger.removeHandler(_hdlr)

    _hdlr = handler
    addLoggingHandler(_hdlr)


def setLogFile(filename):
    #
    # log to rotating files
    #
    hdlr = RotatingFileHandler(filename, "a", 1048576, 5)  # 1 MB by file, 5 files max.
    hdlr.setFormatter(_hwr_formatter)

    setLoggingHandler(hdlr)



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__credits__ = ["EMBL Hamburg"]
__version__ = "2.3."
__category__ = "General"


class EMBLBeamstop(Device, AbstractMotor):
    def __init__(self, name):
        super().__init__(name)

        self.distance = None
        self.default_size = None
        self.default_distance = None
        self.default_direction = None

        self.chan_distance = None
        self.chan_position = None

    def init(self):
        """Reads parameters from xml and adds necessary channels"""
        self.default_size = self.get_property("defaultBeamstopSize")
        self.default_distance = self.get_property("defaultBeamstopDistance")
        self.default_direction = self.get_property("defaultBeamstopDirection")

        if self.default_distance is None:
            self.chan_distance = self.get_channel_object("BeamstopDistance")
            if self.chan_distance is not None:
                self.chan_distance.connect_signal("update", self.distance_changed)
            self.distance_changed(self.chan_distance.get_value())
        else:
            self.distance = float(self.default_distance)

        self.chan_position = self.get_channel_object("BeamstopPosition")

    def is_ready(self):
        """Returns True if device ready"""
        return True

    def distance_changed(self, value):
        """Updates beam stop distance value

        :param value: beamstop distance
        :type value: float
        :return: None
        """
        self.distance = value
        self.emit("beamstopDistanceChanged", value)

    def get_size(self):
        """Returns default beamstop size"""
        return self.default_size

    def set_distance(self, distance):
        """Sets beamstop distance

        :param distance: beamstop distance
        :type distance: float (mm)
        """
        if self.chan_distance is not None:
            self.chan_distance.set_value(distance)
            self.distance_changed(distance)

    def get_distance(self):
        """Returns beamstop distance in mm"""
        return self.distance

    def get_direction(self):
        """Returns beamstop direction"""
        return self.default_direction

    def get_value(self):
        """Returns beamstop position"""
        return self.chan_position.get_value()

    def set_position(self, position):
        """Sets position

        :param position: beamstop position
        :type position: str
        """
        self.chan_position.set_value(position)

    def re_emit_values(self):
        """Reemits available signals"""
        self.emit("beamstopDistanceChanged", self.distance)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


import os
import time

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractCollect import AbstractCollect
from mxcubecore.TaskUtils import task

__credits__ = ["MXCuBE collaboration"]


class CollectMockup(AbstractCollect):
    """ """

    def __init__(self, name):
        """

        :param name: name of the object
        :type name: string
        """

        AbstractCollect.__init__(self, name)

        self.aborted_by_user = False

    def init(self):
        """Main init method"""

        AbstractCollect.init(self)

        self.emit("collectConnected", (True,))
        self.emit("collectReady", (True,))

    def data_collection_hook(self):
        """Main collection hook"""
        self.emit("collectStarted", (None, 1))
        self.emit("fsmConditionChanged", "data_collection_started", True)
        self._store_image_in_lims_by_frame_num(1)
        number_of_images = self.current_dc_parameters["oscillation_sequence"][0][
            "number_of_images"
        ]
        for image in range(
            self.current_dc_parameters["oscillation_sequence"][0]["number_of_images"]
        ):
            if self.aborted_by_user:
                self.ready_event.set()
                self.aborted_by_user = False
                return

            # Uncomment to test collection failed
            # if image == 5:
            #    self.emit("collectOscillationFailed", (self.owner, False,
            #       "Failed on 5", self.current_dc_parameters.get("collection_id")))
            #    self.ready_event.set()
            #    return

            time.sleep(
                self.current_dc_parameters["oscillation_sequence"][0]["exposure_time"]
            )
            self.emit("collectImageTaken", image)
            self.emit("progressStep", (int(float(image) / number_of_images * 100)))
        self.emit_collection_finished()

    def emit_collection_finished(self):
        """Collection finished behaviour"""
        if self.current_dc_parameters["experiment_type"] != "Collect - Multiwedge":
            self._update_data_collection_in_lims()

            last_frame = self.current_dc_parameters["oscillation_sequence"][0][
                "number_of_images"
            ]
            if last_frame > 1:
                self._store_image_in_lims_by_frame_num(last_frame)
            if (
                self.current_dc_parameters["experiment_type"] in ("OSC", "Helical")
                and self.current_dc_parameters["oscillation_sequence"][0]["overlap"]
                == 0
                and last_frame > 19
            ):
                self.trigger_auto_processing("after", self.current_dc_parameters, 0)

        success_msg = "Data collection successful"
        # self.current_dc_parameters["status"] = success_msg
        self.emit(
            "collectOscillationFinished",
            (
                None,
                True,
                success_msg,
                self.current_dc_parameters.get("collection_id"),
                None,
                self.current_dc_parameters,
            ),
        )
        self.emit("collectEnded", None, success_msg)
        self.emit("collectReady", (True,))
        self.emit("progressStop", ())
        self.emit("fsmConditionChanged", "data_collection_successful", True)
        self.emit("fsmConditionChanged", "data_collection_started", False)
        self._collecting = False
        self.ready_event.set()

    def _store_image_in_lims_by_frame_num(self, frame, motor_position_id=None):
        """
        Descript. :
        """
        self.trigger_auto_processing("image", self.current_dc_parameters, frame)
        image_id = self._store_image_in_lims(frame)
        return image_id

    def trigger_auto_processing(self, process_event, frame_number):
        """
        Descript. :
        """
        if HWR.beamline.offline_processing is not None:
            HWR.beamline.offline_processing.execute_autoprocessing(
                process_event,
                self.current_dc_parameters,
                frame_number,
                self.run_offline_processing,
            )

    def stop_collect(self):
        """
        Descript. :
        """
        AbstractCollect.stop_collect(self)
        self.aborted_by_user = True

    @task
    def _take_crystal_snapshot(self, filename):
        HWR.beamline.sample_view.save_snapshot(filename)

    @task
    def _take_crystal_animation(self, animation_filename, duration_sec=1):
        """Rotates sample by 360 and composes a gif file
        Animation is saved as the fourth snapshot
        """
        HWR.beamline.sample_view.save_scene_animation(animation_filename, duration_sec)

    # @task
    # def move_motors(self, motor_position_dict):
    #     """
    #     Descript. :
    #     """
    #     return

    @task
    def move_motors(self, motor_position_dict):
        # TODO We copy, as dictionary is reset in move_motors. CLEAR UP!!
        # TODO clear up this confusion between move_motors and moveMotors
        HWR.beamline.diffractometer.move_motors(motor_position_dict.copy())

    def prepare_input_files(self):
        """
        Descript. :
        """
        i = 1
        while True:
            xds_input_file_dirname = "xds_%s_%s_%d" % (
                self.current_dc_parameters["fileinfo"]["prefix"],
                self.current_dc_parameters["fileinfo"]["run_number"],
                i,
            )
            xds_directory = os.path.join(
                self.current_dc_parameters["fileinfo"]["process_directory"],
                xds_input_file_dirname,
            )
            if not os.path.exists(xds_directory):
                break
            i += 1

        mosflm_input_file_dirname = "mosflm_%s_run%s_%d" % (
            self.current_dc_parameters["fileinfo"]["prefix"],
            self.current_dc_parameters["fileinfo"]["run_number"],
            i,
        )
        mosflm_directory = os.path.join(
            self.current_dc_parameters["fileinfo"]["process_directory"],
            mosflm_input_file_dirname,
        )

        return xds_directory, mosflm_directory, ""

    # rhfogh Added to improve interaction with UI and persistence of values
    def set_wavelength(self, wavelength):
        HWR.beamline.energy.set_wavelength(wavelength)

    def set_energy(self, energy):
        HWR.beamline.energy.set_value(energy)

    def set_transmission(self, transmission):
        HWR.beamline.transmission.set_value(transmission)

    def get_undulators_gaps(self):
        return {"u29": 10}



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""
LNLSBeam class
"""

__copyright__ = """ Copyright © 2016 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"

import ast

from mxcubecore.HardwareObjects.abstract.AbstractBeam import (
    AbstractBeam,
    BeamShape,
)


class LNLSBeam(AbstractBeam):
    def __init__(self, name):
        AbstractBeam.__init__(self, name)

        self._beam_size_dict["slits"] = [9999, 9999]
        self._beam_size_dict["aperture"] = [9999, 9999]
        self._beam_position_on_screen = [1024 / 2, 1024 / 2]
        self._beam_divergence = (0, 0)

    def init(self):
        self._aperture = self.get_object_by_role("aperture")
        if self._aperture is not None:
            self.connect(
                self._aperture,
                "diameterIndexChanged",
                self.aperture_diameter_changed,
            )

            ad = self._aperture.get_diameter_size() / 1000.0
            self._beam_size_dict["aperture"] = [ad, ad]

        self._slits = self.get_object_by_role("slits")
        if self._slits is not None:
            self.connect(self._slits, "valueChanged", self.slits_gap_changed)

            sx, sy = self._slits.get_gaps()
            self._beam_size_dict["slits"] = [sx, sy]

        beam_pos = self.getProperty("default_beam_pos_on_screen")
        self._beam_position_on_screen = list(ast.literal_eval(beam_pos))

        self.evaluate_beam_info()
        self.emit_beam_info_change()
        self.emit("beamPosChanged", (self._beam_position_on_screen,))

    def aperture_diameter_changed(self, name, size):
        """
        Method called when the aperture diameter changes
        Args:
            name (str): diameter name - not used.
            size (float): diameter size in microns
        """
        self._beam_size_dict["aperture"] = [size, size]
        self.evaluate_beam_info()
        self.emit_beam_info_change()

    def slits_gap_changed(self, size):
        """
        Method called when the slits gap changes
        Args:
            size (tuple): two floats indicates beam size in microns
        """
        self._beam_size_dict["slits"] = size
        self.evaluate_beam_info()
        self.emit_beam_info_change()

    def set_beam_position_on_screen(self, beam_x, beam_y):
        """
        Sets beam mark position on screen
        #TODO move method to sample_view
        Args:
            beam_x (int): horizontal position in pixels
            beam_y (int): vertical position in pixels
        """
        self._beam_position_on_screen = (beam_x, beam_y)
        self.emit("beamPosChanged", (self._beam_position_on_screen,))

    def get_slits_gap(self):
        """
        Returns: tuple with beam size in microns
        """
        self.evaluate_beam_info()
        return self._beam_size_dict["slits"]

    def set_slits_gap(self, width_microns, height_microns):
        """
        Sets slits gap in microns
        Args:
            width_microns (int):
            height_microns (int):
        """
        if self._slits:
            self._slits.set_horizontal_gap(width_microns / 1000.0)
            self._slits.set_vertical_gap(height_microns / 1000.0)

    def get_aperture_pos_name(self):
        """
        Returns (str): name of current aperture position
        """
        if self._aperture:
            return self._aperture.get_current_pos_name()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import logging

import gevent
from PyQt5.QtCore import (
    QObject,
    pyqtSignal,
)

from mxcubecore.HardwareObjects.abstract.AbstractMachineInfo import AbstractMachineInfo
from mxcubecore.HardwareObjects.TangoMachineInfo import TangoMachineInfo


class P11MachineInfo(TangoMachineInfo, QObject):
    # Declare the signal
    valuesChanged = pyqtSignal(dict)

    # Declare the signal
    machineCurrentChanged = pyqtSignal(float, bool)

    def __init__(self, *args, **kwargs):
        """Initializes the P11MachineInfo and QObject."""
        # Initialize both parent classes explicitly
        TangoMachineInfo.__init__(
            self, *args, **kwargs
        )  # Initialize the TangoMachineInfo class
        QObject.__init__(self)  # Initialize the QObject class

    def init(self):
        """Initialize Tango channels for machine info."""
        super().init()

        self.emit_values()  # Emit initial values after init
        gevent.spawn(self.periodic_update)

    def periodic_update(self):
        while True:
            self.emit_values()
            gevent.sleep(3)

    def emit_values(self):
        """Emit the current machine info values."""
        values_dict = {
            "current": {
                "title": "Current (mA)",
                "value": self.get_current() or 0,  # Ensure value is not None
            },
            "lifetime": {"title": "Lifetime (h)", "value": self.get_lifetime() or 0},
            "energy": {
                "title": "Energy (GeV)",
                "value": self.get_maschine_energy() or 0,
            },
            "message": {
                "title": "Message",
                "value": self.get_message() or "No message",
            },
        }

        logging.info(f"Emitting machine info values: {values_dict}")
        self.valuesChanged.emit(values_dict)  # Emit the valuesChanged signal

    def get_value(self):
        """Returns a dictionary of the current machine information."""
        try:
            self._mach_info_dict = {}  # Initialize dictionary

            # Fetch the current values from their respective methods
            self._mach_info_dict["current"] = self.get_current() or 0
            self._mach_info_dict["lifetime"] = self.get_lifetime() or 0
            self._mach_info_dict["energy"] = self.get_maschine_energy() or 0
            self._mach_info_dict["message"] = self.get_message() or "No message"

            logging.info(
                f"Machine Info Dictionary: {self._mach_info_dict}"
            )  # Log the populated dictionary
            return self._mach_info_dict  # Return the dictionary with machine info

        except Exception as e:
            logging.error(f"Error getting machine values: {e}")
            return {}  # Return empty dictionary if there's an error

    def get_current(self):
        try:
            current_value = (
                self.current.get_value()
            )  # Fetch the value from the Tango hardware
            return round(current_value, 2)
        except AttributeError:
            logging.error("Error reading 'current': Attribute 'current' not found.")
            return 0

    def get_lifetime(self):
        try:
            lifetime_value = self.lifetime.get_value()  # Fetch lifetime value
            return round(lifetime_value, 2)
        except AttributeError:
            logging.error("Error reading 'lifetime': Attribute 'lifetime' not found.")
            return 0

    def get_maschine_energy(self):
        try:
            energy_value = self.energy.get_value()  # Fetch energy value
            return round(energy_value, 2)
        except AttributeError:
            logging.error("Error reading 'energy': Attribute 'energy' not found.")
            return 0

    def get_message(self):
        try:
            message_value = self.message.get_value()  # Fetch message from hardware
            return message_value
        except AttributeError:
            logging.error("Error reading 'message': Attribute 'message' not found.")
            return "Message unavailable"

    def update_current(self, current_value, in_range):
        # Emit the signal with the current machine value and its range status
        self.machineCurrentChanged.emit(current_value, in_range)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import ast
from enum import Enum

from mxcubecore.HardwareObjects.abstract.AbstractNState import (
    AbstractNState,
    BaseValueEnum,
)

redis_flag = False
try:
    import redis

    redis_flag = True
except:
    print("impossible to import redis")


class P11Zoom(AbstractNState):
    def __init__(self, name):
        self.camera_hwobj = None
        self.pixels_per_mm = None
        self.closest_zoom = None
        self.redis = redis.StrictRedis()

        super().__init__(name)

    def init(self):
        self._pixels_per_mm = self.get_property("pixels_per_mm")
        self._overview_pixels_per_mm = self.get_property("overview_pixels_per_mm")
        self.camera_hwobj = self.get_object_by_role("camera")

        self.connect(self.camera_hwobj, "zoomChanged", self.zoom_value_changed)

        if self.camera_hwobj is None:
            self.log.debug("P11Zoom.py - cannot connect to camera hardware object")

        AbstractNState.init(self)

    def initialise_values(self):
        self.VALUES = Enum("ZoomEnum", ast.literal_eval(self.get_property("values")))

    def get_state(self):
        return self.STATES.READY

    def is_moving(self):
        return False

    def get_value(self):
        return self.get_zoom()

    def _set_value(self, value):
        self.set_zoom(value)

    def get_pixels_per_mm(self):
        current_zoom = self.get_zoom()

        if current_zoom == 0:
            px_per_mm = self._overview_pixels_per_mm
        else:
            self.log.debug(
                "getting pixels per mm: %s - %s"
                % (str(self._pixels_per_mm), str(current_zoom))
            )
            px_per_mm = self._pixels_per_mm * current_zoom

        scale = self.camera_hwobj.get_scale()
        pixels_per_mm = [int(px_per_mm * scale), int(px_per_mm * scale)]

        if redis_flag:
            print("saving calibration to redis", pixels_per_mm)
            self.redis.set("pixels_per_mm", str(pixels_per_mm))

        return pixels_per_mm

    def get_zoom(self):
        zoom = self.camera_hwobj.get_zoom()
        if zoom is not None:
            self.current_value = zoom
        return self.current_value

    def set_zoom_value(self, value):
        self.camera_hwobj.set_zoom(value)

    def set_zoom(self, zoom):
        self.camera_hwobj.set_zoom(zoom.value)

    def get_current_zoom(self):
        self.get_zoom()
        self.update_zoom()
        return self.closest_zoom, self.get_zoom()

    def zoom_value_changed(self, value):
        print("zoom_value_changed")
        self.current_value = value
        self.update_value(value)
        self.update_zoom()
        # self.get_pixels_per_mm()

    def update_zoom(self):
        dist = None
        value = self.get_value()

        for zoom in self.VALUES:
            if value == 0:
                self.closest_zoom = zoom
                break

            _dist = abs(value - zoom.value)
            if dist is None or _dist < dist:
                dist = _dist
                self.closest_zoom = zoom

        if self.closest_zoom is not None:
            self.emit("predefinedPositionChanged", (self.closest_zoom, value))
        else:
            self.emit("predefinedPositionChanged", (None, None))



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Convert exporter states to HardwareObject states
"""

from enum import Enum

from mxcubecore.BaseHardwareObjects import HardwareObjectState

__copyright__ = """ Copyright © 2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class ExporterStates(Enum):
    """Convert exporter states to HardwareObject amd Motor states"""

    UNKNOWN = HardwareObjectState.UNKNOWN
    STANDBY = HardwareObjectState.WARNING
    DISABLE = HardwareObjectState.WARNING
    ALARM = HardwareObjectState.WARNING
    BUSY = HardwareObjectState.BUSY
    MOVING = HardwareObjectState.BUSY
    RUNNING = HardwareObjectState.BUSY
    INITIALIZING = HardwareObjectState.BUSY
    READY = HardwareObjectState.READY
    ON = HardwareObjectState.READY
    OFF = HardwareObjectState.READY
    CLOSED = HardwareObjectState.READY
    OPEN = HardwareObjectState.READY
    FAULT = HardwareObjectState.FAULT
    INVALID = HardwareObjectState.FAULT
    OFFLINE = HardwareObjectState.FAULT



import errno
import json
import logging
import os
import time

import gevent
from devtools import debug
from pydantic.v1 import (
    BaseModel,
    Field,
)
from typing_extensions import (
    Literal,
    Union,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.model.common import (
    CommonCollectionParamters,
    ISPYBCollectionParameters,
    LegacyParameters,
    PathParameters,
    StandardCollectionParameters,
)
from mxcubecore.queue_entry.base_queue_entry import BaseQueueEntry

DEFAULT_MAX_FREQ = 925


class MXPathParameters(PathParameters):
    use_experiment_name: bool = Field(
        False, description="Whether to use the experiment name in the data path"
    )


class BaseUserCollectionParameters(BaseModel):
    exp_time: float = Field(95e-6, gt=0, lt=1, description="s")


class MXBaseQueueTaskParameters(BaseModel):
    path_parameters: MXPathParameters
    common_parameters: CommonCollectionParamters
    collection_parameters: StandardCollectionParameters
    legacy_parameters: LegacyParameters
    lims_parameters: Union[ISPYBCollectionParameters, None]

    def update_dependent_fields(field_data):
        return {}

    @staticmethod
    def ui_schema():
        return json.dumps(
            {
                "ui:order": [
                    "num_images",
                    "exp_time",
                    "osc_range",
                    "osc_start",
                    "resolution",
                    "transmission",
                    "energy",
                    "*",
                ],
                "ui:submitButtonOptions": {
                    "norender": "true",
                },
            }
        )


class MXBaseQueueEntry(BaseQueueEntry):
    """
    Defines common MX collection methods.
    """

    def __init__(self, view, data_model):
        super().__init__(view=view, data_model=data_model)
        self._beamline_values = None

    def get_data_directory(self):
        return self.get_data_model().get_path_template().directory

    def get_data_path(self):
        return self.get_data_model().get_path_template().get_image_path()

    def _check_file(self, fname, wait_time=60):
        if not os.path.isfile(fname):
            logging.getLogger("HWR").info("File {fname} not yet written")
            logging.getLogger("HWR").info("Checking for file {fname}...")

            for i in range(1, 3):
                time.sleep(60)

                if os.path.isfile(fname):
                    break

                logging.getLogger("HWR").info(
                    "{fname} still not found after {3-i} retries"
                )

            logging.getLogger("HWR").warning("File {fname} not found")

    def monitor_progress(self):
        num_images = self._data_model._task_data.collection_parameters.num_images
        exp_time = self._data_model._task_data.collection_parameters.exp_time
        images_per_file = HWR.beamline.detector.images_per_file
        num_files = num_images // images_per_file
        dp = 100 / num_files
        total_progress = 0

        pt = self.get_data_model().get_path_template()
        h5_master_path = pt.get_image_file_name()

        for i in range(1, num_files):
            time.sleep(images_per_file * exp_time)

            current_file = pt.get_actual_file_path(
                h5_master_path, i * images_per_file - 1
            )

            gevent.spawn(self._check_file, current_file)

            total_progress += dp
            self.emit_progress(total_progress)

    def create_directory(self):
        try:
            os.makedirs(self.get_data_directory())
        except os.error as e:
            if e.errno != errno.EEXIST:
                raise

    def start_processing(self, exp_type):
        data_root_path = self.get_data_path()

    def prepare_acquisition(self):
        pass

    def execute(self):
        super().execute()
        debug(self._data_model._task_data)

    def pre_execute(self):
        super().pre_execute()
        self.create_directory()

        self.emit_progress(0)

    def post_execute(self):
        super().post_execute()
        self.emit_progress(1)

    def emit_progress(self, progress):
        HWR.beamline.collect.emit_progress(progress)

    def stop(self):
        super().stop()
        HWR.beamline.detector.stop_acquisition()
